__all__ = ('CharClass', 'Barbarian', 'Bard', 'Cleric', 'Druid', 'Fighter',
           'Monk', 'Paladin', 'Ranger', 'Rogue', 'Sorceror', 'Warlock',
           'Wizard', 'RevisedRanger', 'available_classes')

from .classes import CharClass
from .barbarian import Barbarian
from .bard import Bard
from .cleric import Cleric
from .druid import Druid
from .fighter import Fighter
from .monk import Monk
from .paladin import Paladin
from .ranger import (Ranger, RevisedRanger)
from .rogue import Rogue
from .sorceror import Sorceror
from .warlock import Warlock
from .wizard import Wizard

available_classes = [Barbarian, Bard, Cleric, Druid, Fighter, Monk, Paladin,
                     Ranger, Rogue, Sorceror, Warlock, Wizard, RevisedRanger]



import math
from collections import namedtuple
from .armor import NoArmor, NoShield, HeavyArmor, Shield, Armor
from .weapons import Weapon
from .features import (UnarmoredDefenseMonk, UnarmoredDefenseBarbarian,
                       DraconicResilience, Defense, FastMovement,
                       UnarmoredMovement, GiftOfTheDepths, RemarkableAthelete,
                       SeaSoul, JackOfAllTrades, SoulOfTheForge, QuickDraw,
                       NaturalExplorerRevised, FeralInstinct, DreadAmbusher,
                       SuperiorMobility, AmbushMaster, RakishAudacity,
                       NaturalArmor)
from math import ceil


def findattr(obj, name):
    """Similar to builtin getattr(obj, name) but more forgiving to
    whitespace and capitalization.
    
    """
    # Come up with several options
    name = name.strip()
    # check for +X weapons, armor, shields
    bonus = 0
    for i in range(1, 11):
        if (f'+{i}' in name) or (f'+ {i}' in name):
            bonus = i
            name = name.replace(f'+{i}', '').replace(f'+ {i}', '')
            break
    py_name = name.replace('-', '_').replace(' ', '_').replace("'", "")
    camel_case = "".join([s.capitalize() for s in py_name.split('_')])
    if hasattr(obj, py_name):
        # Direct lookup
        attr = getattr(obj, py_name)
    elif hasattr(obj, camel_case):
        # CamelCase lookup
        attr = getattr(obj, camel_case)
    else:
        raise AttributeError(f'{obj} has no attribute {name}')
    if bonus > 0:
        if issubclass(attr, Weapon) or issubclass(attr, Shield) or issubclass(attr, Armor):
            attr = attr.improved_version(bonus)
    return attr


def mod_str(modifier):
    """Converts a modifier to a string, eg 2 -> '+2'."""
    return '{:+d}'.format(modifier)
    if modifier == 0:
        return str(modifier)
    else:
        return '{:+}'.format(modifier)


AbilityScore = namedtuple('AbilityScore',
                          ('value', 'modifier', 'saving_throw'))


class Ability():
    ability_name = None
    
    def __init__(self, default_value=10):
        self.default_value = default_value
    
    def __set_name__(self, character, name):
        self.ability_name = name
    
    def _check_dict(self, obj):
        if not hasattr(obj, '_ability_scores'):
            # No ability score dictionary exists
            obj._ability_scores = {
                self.ability_name: self.default_value
            }
        elif self.ability_name not in obj._ability_scores.keys():
            # ability score dictionary exists but doesn't have this ability
            obj._ability_scores[self.ability_name] = self.default_value
    
    def __get__(self, character, Character):
        self._check_dict(character)
        score = character._ability_scores[self.ability_name]
        modifier = math.floor((score - 10) / 2)
        # Check for proficiency
        saving_throw = modifier
        if self.ability_name is not None and hasattr(character, 'saving_throw_proficiencies'):
            is_proficient = (self.ability_name in character.saving_throw_proficiencies)
            if is_proficient:
                saving_throw += character.proficiency_bonus
        # Create the named tuple
        value = AbilityScore(modifier=modifier, value=score, saving_throw=saving_throw)
        return value
    
    def __set__(self, character, val):
        self._check_dict(character)
        character._ability_scores[self.ability_name] = val
        self.value = val


class Skill():
    """An ability-based skill, such as athletics."""
    
    def __init__(self, ability):
        self.ability_name = ability
    
    def __set_name__(self, character, name):
        self.skill_name = name.lower().replace('_', ' ')
        self.character = character
    
    def __get__(self, character, owner):
        ability = getattr(character, self.ability_name)
        modifier = ability.modifier
        # Check for proficiency
        is_proficient = self.skill_name in character.skill_proficiencies
        if is_proficient:
            modifier += character.proficiency_bonus
        elif character.has_feature(JackOfAllTrades):
            modifier += character.proficiency_bonus // 2
        elif character.has_feature(RemarkableAthelete):
            if self.ability_name.lower() in ('strength',
                                             'dexterity', 'constitution'):
                modifier += ceil(character.proficienc_bonus / 2.)
        
        # Check for expertise
        is_expert = self.skill_name in character.skill_expertise
        if is_expert:
            modifier += character.proficiency_bonus
        return modifier


class ArmorClass():
    """
    The Armor Class of a character
    """

    def __get__(self, char, Character):
        armor = char.armor or NoArmor()
        ac = armor.base_armor_class
        # calculate and apply modifiers
        if armor.dexterity_mod_max is None:
            ac += char.dexterity.modifier
        else:
            ac += min(char.dexterity.modifier, armor.dexterity_mod_max)
        if char.has_feature(NaturalArmor):
            ac = max(ac, 13 + char.dexterity.modifier)
        shield = char.shield or NoShield()
        ac += shield.base_armor_class
        # Compute feature-specific additions
        if char.has_feature(UnarmoredDefenseMonk):
            if (isinstance(armor, NoArmor) and isinstance(shield, NoShield)):
                ac += char.wisdom.modifier
        if char.has_feature(UnarmoredDefenseBarbarian):
            if isinstance(armor, NoArmor):
                ac += char.constitution.modifier
        if char.has_feature(DraconicResilience):
            if isinstance(armor, NoArmor):
                ac += 3
        if char.has_feature(Defense):
            if not isinstance(armor, NoArmor):
                ac += 1
        if char.has_feature(SoulOfTheForge):
            if isinstance(armor, HeavyArmor):
                ac += 1
        # Check if any magic items add to AC
        for mitem in char.magic_items:
            if hasattr(mitem, 'ac_bonus'):
                ac += mitem.ac_bonus
        return ac
        

class Speed():
    """
    The speed of a character
    """

    def __get__(self, char, Character):
        speed = char.race.speed
        other_speed = ''
        if isinstance(speed, str):
            other_speed = speed[2:]
            speed = int(speed[:2])  # ignore other speeds, like fly
        if char.has_feature(FastMovement):
            if not isinstance(char.armor, HeavyArmor):
                speed += 10
        if char.has_feature(SuperiorMobility):
            speed += 10
        if isinstance(char.armor, NoArmor) or (char.armor is None):
            for f in char.features:
                if isinstance(f, UnarmoredMovement):
                    speed += f.speed_bonus
        if char.has_feature(GiftOfTheDepths):
            if 'swim' not in other_speed:
                other_speed += ' ({:d} swim)'.format(speed)
        if char.has_feature(SeaSoul):
            if 'swim' not in other_speed:
                other_speed += ' (30 swim)'
        return '{:d}{:s}'.format(speed, other_speed)


class Initiative():
    """A character's initiative"""
    def __get__(self, char, Character):
        ini = char.dexterity.modifier
        if char.has_feature(QuickDraw):
            ini += char.proficiency_bonus
        if char.has_feature(DreadAmbusher):
            ini += char.wisdom.modifier
        if char.has_feature(RakishAudacity):
            ini += char.charisma.modifier
        ini = '{:+d}'.format(ini)
        has_advantage = (char.has_feature(NaturalExplorerRevised) or
                         char.has_feature(FeralInstinct) or
                         char.has_feature(AmbushMaster))
        if has_advantage:
            ini += '(A)'
        return ini
    



import json
import re

OUTFILE = 'spells_new.py'
INFILE = '/home/mwolf/Documents/dungeons_and_dragons/spells.json'

data = json.load(open(INFILE, mode='r'))

components_re = re.compile('([VSM])?[, ]*([VSM])?[, ]*([VSM])?[, ]*'
                           '(?:\(([-a-zA-Z ,.0-9;’—()]+)\))?')

def parse_components(string):
    result = components_re.match(string)
    components = tuple(r for r in result.groups()[0:3] if r is not None)
    materials = result.groups()[3]
    if 'M' in components:
        assert materials is not None
    if materials is None:
        materials = ''
    return components, materials

print(parse_components('V'))
print(parse_components('V, S'))
print(parse_components('S, V'))
print(parse_components("V, S, M (a tiny strip of white cloth)"))

with open(OUTFILE, mode='w') as out:

    for spell_name, spell in data.items():
        # Read the components list to determine which components there are
        class_name = ''.join([s.capitalize() for s in spell_name.split()])
        # Process the components
        try:
            components, materials = parse_components(spell["components"])
        except AssertionError:
            print(spell_name, spell)
        new_s = (f'class {class_name}(Spell):\n'
                 f'    """{spell["description"]}"""\n'
                 f'    name = "{spell_name}"\n'
                 f'    level = {spell["level"]}\n'
                 f'    casting_time = "{spell["casting_time"]}"\n'
                 f'    components = {components}\n'
                 f'    materials = "{materials}"\n'
                 f'    duration = "{spell["duration"]}"\n'
                 f'    magic_school = "{spell["school"].capitalize()}"\n'
                 f'    classes = ()\n'
                 '\n\n')
        # print(new_s)
        out.write(new_s)



from . import features as feats


class Background():
    name = "Generic background"
    owner = None
    skill_proficiencies = ()
    weapon_proficiencies = ()
    proficiencies_text = ()
    skill_choices = ()
    num_skill_choices = 0
    features = ()
    languages = ()

    def __init__(self, owner=None):
        self.owner = owner
        cls = type(self)
        self.features = tuple([f(owner=self.owner) for f in cls.features])

    def __str__(self):
        return self.name


class Acolyte(Background):
    name = "Acolyte"
    skill_proficiencies = ('insight', 'religion')
    languages = ("[choose one]", "[choose one]")
    features = (feats.ShelterOfTheFaithful,)


class Charlatan(Background):
    name = "Charlatan"
    skill_proficiencies = ('deception', 'sleight of hand')
    features = (feats.FalseIdentity,)


class Criminal(Background):
    name = "Criminal"
    skill_proficiencies = ('deception', 'stealth')
    features = (feats.CriminalContact,)
    

class Spy(Criminal):
    name = "Spy"


class Entertainer(Background):
    name = "Entertainer"
    skill_proficiencies = ('acrobatics', 'performance')
    features = (feats.ByPopularDemand,)


class Gladiator(Entertainer):
    name = "Gladiator"
    

class FolkHero(Background):
    name = "Folk Hero"
    skill_proficiencies = ('animal handling', 'survival')
    features = (feats.RusticHospitality,)


class GuildArtisan(Background):
    name = "Guild Artisan"
    skill_proficiencies = ('insight', 'persuasion')
    languages = ("[choose one]", "[choose one]")
    features = (feats.GuildMembership,)


class GuildMerchant(GuildArtisan):
    name = "Guild Merchant"


class Hermit(Background):
    name = "Hermit"
    skill_proficiencies = ("medicine", "religion")
    languages = ("[choose one]", )
    features = (feats.Discovery,)


class Noble(Background):
    name = "Noble"
    skill_proficiencies = ("history", 'persuasion')
    languages = ("[choose one]", )
    features = (feats.PositionOfPrivilege,)


class Knight(Noble):
    name = "Knight"


class Outlander(Background):
    name = "Outlander"
    skill_proficiencies = ('athletics', 'survival')
    languages = ("[choose one]", )
    features = (feats.Wanderer,)
    

class Sage(Background):
    name = "Sage"
    skill_proficiencies = ('arcana', 'history')
    languages = ("[choose one]", '[choose one]')
    features = (feats.Researcher,)


class Sailor(Background):
    name = "Sailor"
    skill_proficiencies = ('athletics', 'perception')
    features = (feats.ShipsPassage,)


class Pirate(Sailor):
    name = "Pirate"


class Soldier(Background):
    name = "Soldier"
    skill_proficiencies = ('athletics', 'intimidation')
    features = (feats.MilitaryRank,)


class Urchin(Background):
    name = "Urchin"
    skill_proficiencies = ('sleight of hand', 'stealth')
    features = (feats.CitySecrets,)


# Sword's Coast Adventurers Guide
class CityWatch(Background):
    name = "City Watch"
    skill_proficiencies = ('athletics', 'insight')
    languages = ('[choose one]', '[choose one]')
    features = (feats.WatchersEye,)


class ClanCrafter(Background):
    name = "Clan Crafter"
    skill_proficiencies = ('history', 'insight')
    languages = ('Dwarvish',)
    features = (feats.RespectOfTheStoutFolk,)


class CloisteredScholar(Background):
    name = "Cloistered Scholar"
    skill_proficiencies = ('history',)
    skill_choices = ('arcana', 'nature', 'religion')
    num_skill_choices = 1
    languages = ('[choose one]', '[choose one]')
    features = (feats.LibraryAccess,)


class Courtier(Background):
    name = "Courtier"
    skill_proficiencies = ("insight", 'persuasion')
    languages = ('[choose one]', '[choose one]')
    features = (feats.CourtFunctionary,)


class FactionAgent(Background):
    name = "Faction Agent"
    skill_proficiencies = ('insight',)
    skill_choices = ('animal handling', 'arcana', 'deception',
                     'history', 'intimidation', 'investigation',
                     'medicine', 'nature', 'perception', 'performance',
                     'persuasion', 'religion', 'survival')
    num_skill_choices = 1
    languages = ('[choose one]', '[choose one]')
    features = (feats.SafeHaven,)


class FarTraveler(Background):
    name = 'Far Traveler'
    skill_proficiencies = ('insight', 'perception')
    languages = ('[choose one]',)
    features = (feats.AllEyesOnYou,)


class Inheritor(Background):
    name = "Inheritor"
    skill_proficiencies = ('survival',)
    skill_choices = ('arcana', 'history', 'religion')
    num_skill_choices = 1
    languages = ('[choose one]',)
    features = (feats.Inheritance,)


class KnightOfTheOrder(Background):
    name = "Knight of the Order"
    skill_proficiencies = ('persuasion',)
    skill_choices = ('arcana', 'history', 'nature', 'religion')
    num_skill_choices = 1
    languages = ('[choose one]',)
    features = (feats.KnightlyRegard,)


class MercenaryVeteran(Background):
    name = "Mercenary Veteran"
    skill_proficiencies = ('athletics', 'persuasion')
    features = (feats.MercenaryLife,)
    
    
class UrbanBountyHunter(Background):
    name = 'Urban Bounty Hunter'
    skill_proficiencies = ()
    skill_choices = ('Deception', 'Insight', 'Persuasion', 'Stealth')
    num_skill_choices = 2
    features = (feats.EarToTheGround,)


class UthgardtTribeMember(Background):
    name = "Uthgardt Tribe Member"
    skill_profifiencies = ('athletics', 'survival')
    languages = ('[choose one]',)
    features = (feats.UthgardtHeritage,)


class WaterdhavianNoble(Background):
    name = "Waterdhavian Noble"
    skill_proficiencies = ('history', 'persuasion')
    languages = ('[choose one]',)
    features = (feats.KeptInStyle,)
    
    
PHB_backgrounds = [Acolyte, Charlatan, Criminal, Spy, Entertainer,
                   Gladiator, FolkHero, GuildArtisan, GuildMerchant,
                   Hermit, Noble, Knight, Outlander, Sage, Sailor,
                   Pirate, Soldier, Urchin]

SCAG_backgrounds = [CityWatch, ClanCrafter, CloisteredScholar, Courtier,
                    FactionAgent, FarTraveler, Inheritor, KnightOfTheOrder,
                    MercenaryVeteran, UrbanBountyHunter, UthgardtTribeMember,
                    WaterdhavianNoble]

available_backgrounds = PHB_backgrounds + SCAG_backgrounds

__all__ = tuple([b.name for b in available_backgrounds]) + (
    'PHB_backgrounds', 'SCAG_backgrounds', 'available_backgrounds')



from .features import Feature


# PHB
class GreatWeaponMaster(Feature):
    """You’ve learned to put the weight of a weapon to your advantage, letting its
    momentum empower your strikes. You gain the following benefits:

    -- On your turn, when you score a critical hit with a melee weapon or
    reduce a creature to 0 hit points with one, you can make one melee weapon
    attack as a bonus action.

    -- Before you make a melee attack with a heavy weapon that you are
    proficient with, you can choose to take a -5 penalty to the attack
    roll. If the attack hits, you add +10 to the attack’s damage

    """
    name = "Great Weapon Master"
    source = "Feats"


class Actor(Feature):
    """Skilled at mimicry and dramatics, you gain the following benefits:

    -- Increase your Charisma score by 1, to a maximum of 20.

    -- You have an advantage on Charisma (Deception) and Charisma (Performance) checks when trying to pass yourself off as a different person.

    --You can mimic the speech of another person or the sounds made by other
    creatures. You must have heard the person speaking, or heard the creature
    make the sound, for at least 1 minute. A successful Wisdom (Insight) check
    contested by your Charisma (Deception) check allows a listener to determine
    that the effect is faked.

    """
    name = "Actor"
    source = "Feats"



class MagicItem():
    """
    Generic Magic Item. Add description here.

    """
    name = ''
    ac_bonus = 0
    requires_attunement = False
    needs_implementation = False
    rarity = ''

    def __init__(self, owner=None):
        self.owner = owner

    def __str__(self):
        return self.name

    def __repr__(self):
        return '\"{:s}\"'.format(str(self))


class RingOfProtection(MagicItem):
    """
    You gain a +1 bonus to AC and Saving Throws while wearing this ring.

    """
    name = "Ring of Protection"
    ac_bonus = 1
    requires_attunement = True
    rarity = 'Rare'


class DecanterOfEndlessWater(MagicItem):
    """This stoppered flask sloshes when shaken, as if it contains water. The
    decanter weighs 2 pounds.

    You can use an action to remove the stopper and speak one of three command
    words, whereupon an amount of fresh water or salt water (your choice) pours
    out of the flask. The water stops pouring out at the start of your next
    turn. Choose from the following options:

    --"Stream" produces 1 gallon of water.

    --"Fountain" produces 5 gallons of water.

    --"Geyser" produces 30 gallons of water that gushes forth in a geyser 30
    feet long and 1 foot wide. As a bonus action while holding the decanter,
    you can aim the geyser at a creature you can see within 30 feet of you. The
    target must succeed on a DC 13 Strength saving throw or take 1d4
    bludgeoning damage and fall prone. Instead of a creature, you can target an
    object that isn't being worn or carried and that weighs no more than 200
    pounds. The object is either knocked over or pushed up to 15 feet away from
    you.

    """
    name = "Decanter of Endless Water"
    rarity = 'Uncommon'


class ToothOfAnimalFriendship(MagicItem):
    """While holding this wolf's tooth, you can expend it's one charge to cast
    Animal Friendship (DC 13) or Speak With Animals. 

    The charge resets at the next Dawn.
    """
    name = "Tooth of Animal Friendship"
    rarity = 'Uncommon'


class CloakOfBillowing(MagicItem):
    """While wearing this cloak, you can use a bonus action to make it billow
    dramatically.

    """
    name = "Cloak of Billowing"
    rarity = "Common"
    

class CapeOfTheMountebank(MagicItem):
    """This cape smells faintly of brimstone. While wearing it, you can use it to
    cast the Dimension Door spell as an action. This property of the cape can't
    be used again until the next dawn.

    When you disappear, you leave behind a cloud of smoke, and you appear in a
    similar cloud of smoke at your destination. The smoke lightly obscures the
    space you left and the space you appear in, and it dissipates at the end of
    your next turn. A light or stronger wind disperses the smoke.

    """
    name = "Cape of the Mountebank"
    rarity = "Rare"
    requires_attunement = True


class EyesOfCharming(MagicItem):
    """These Crystal lenses fit over the eyes. They have 3 Charges. While wearing
    them, you can expend 1 charge as an action to cast the Charm Person spell
    (save DC 13) on a humanoid within 30 feet of you, provided that you and the
    target can see each other. The lenses regain all expended Charges daily at
    dawn.

    """
    name = "Eyes of Charming"
    rarity = "Uncommon"
    requires_attunement = True


class CharlattansDie(MagicItem):
    """Whenever you roll this six—sided die, you can control which number it
    rolls.

    """
    name = "Charlattan's Die"
    rarity = "Common"


class PipeOfSmokeMonsters(MagicItem):
    """While smoking this pipe, you can use an action to ex- hale a puff of smoke
    that takes the form of a single crea— ture, such as a dragon, a flumph, or
    a froghemoth. The form must be small enough to fit in a 1-foot cube and
    loses its shape after a few seconds, becoming an ordi- nary puff of smoke.

    """
    name = 'Pipe of Smoke Monsters'
    rarity = "Common"


class CoinsOfCommunication(MagicItem):
    """This set of multiple coins are virtually indistinguishable from regular Gold
    Pieces, but are connected by magic. Once per day, a holder of any of any
    coin can whisper a single word into it, after which all coins will
    immediately vibrate and the word will replace a word in the traditional
    Kings Message imprinted on the coin. This ability cannot be used again by
    the holder of any of the coins until the following dawn.

    """
    name = "Coins of Communication"
    rarity = "Uncommon"

    
class FlameTongue(MagicItem):
    """You can use a Bonus Action to speak this magic sword's Command Word, causing
    flames to erupt from the blade. These flames shed bright light in a 40-foot
    radius and dim light for an additional 40 feet. While the sword is ablaze,
    it deals an extra 2d6 fire damage to any target it hits. The flames last
    until you use a Bonus Action to speak the Command Word again or until you
    drop or sheathe the sword

    """
    name = "Flame Tongue"
    rarity = "Rare"
    requires_attunement = True


class SpearOfLightning(MagicItem):
    """When you hurl it and speak its Command Word, it transforms into a bolt of
    lightning, forming a line 5 feet wide that extends out from you to a target
    within 120 feet. Each creature in the line excluding you and the target
    must make a DC 13 Dexterity saving throw, taking 4d6 lightning damage on a
    failed save, and half as much damage on a successful one. The Lightning
    Bolt turns back into a spear when it reaches the target.

    Make a ranged weapon Attack against the target. On a hit, the target takes
    damage from the spear plus 4d6 lightning damage.

    The spear's property can't be used again until the next dawn. In the
    meantime, the spear can still be used as a Magic Weapon.
    """
    requires_attunement = True
    name = "Lightning Spear"

    
class AmuletOfTheEel(MagicItem):
    """While holding this amulet, you can breath water and air, and have a
    swimming speed of 20'."""
    name = "Amulet of the Eel"
    

class BracersOfMagnetism(MagicItem):
    """When wearing these bracers, you can use a bonus action to speak their
    command word, causing them to become magnetically attractive to each other.

    While active, the wearer's arms are secured together, requiring a succesful DC 25 
    STR (Athletics) check to separate them by six inches. The wearer has advantage 
    on all STR (Athletics) checks made to grapple, but disadvantage on all
    weapon attacks and DEX (Sleight of Hand) checks. 

    The wearer can use another bonus action to speak the command word and 
    deactivate the magnetic effect.

    The magnetic effect fails if the bracers are more than 10' apart.
    """
    requires_attunement = True
    name = "Bracers of Magnetism"


class ShieldOfFaces(MagicItem):
    """This +1 metallic shield has a detailed face chisled into its front surface.
    As a bonus action, the wielder can use a bonus action to mentally command
    the shield to adopt a new emotional state (such as "smiling", "laughing",
    "crying", etc.)

    """
    requires_attunement = True
    name = "Shield of Faces"


class GlowingSword(MagicItem):
    """
    This strange longsword glows at odd times.
    """
    name = "Glowing Sword"



class CharacterFileFormatError(ValueError):
    """The given file is not a valid Dungeons and Dragons file."""

class DiceError(ValueError):
    """Improper formatting for a dice string."""

class LatexError(OSError):
    """PDFLatex did not execute correctly."""

class LatexNotFoundError(LatexError):
    """PDFLatex did not execute correctly."""    

class MonsterError(AttributeError):
    """Error retriving or using a D&D Monster."""



from .. import (weapons, features, spells)
from .classes import CharClass, SubClass
from collections import defaultdict


# PHB
class Archfey(SubClass):
    """Your patron is a lord or lady of the fey, a creature of legend who holds
    secrets that were forgotten before the mortal races were born. This being’s
    motivations are often inscrutable, and sometimes whimsical, and might
    involve a striving for greater magical power or the settling of age-old
    grudges. Beings of this sort include the Prince of Frost; the Queen of Air
    and Darkness, ruler of the Gloaming Court; Titania of the Summer Court; her
    consort Oberon, the Green Lord; Hyrsam, the Prince of Fools; and ancient
    hags

    """
    name = "The Archfey Patron"
    features_by_level = defaultdict(list)
    features_by_level[1] = [features.FeyPresence]
    features_by_level[6] = [features.MistyEscape]
    features_by_level[10] = [features.BeguilingDefenses]
    features_by_level[14] = [features.DarkDelirium]


class Fiend(SubClass):
    """You have made a pact with a fiend from the lower planes o f existence, a
    being whose aims are evil, even if you strive against those aims. Such
    beings desire the corruption or destruction of all things, ultimately
    including you. Fiends powerful enough to forge a pact include demon lords
    such as Demogorgon, Orcus, Fraz’Urb-luu, and Baphomet; archdevils such as
    Asmodeus, Dispater, Mephistopheles, and Belial; pit fiends and balors that
    are especially mighty; and ultroloths and other lords of the yugoloths

    """
    name = "The Fiend Patron"
    features_by_level = defaultdict(list)
    features_by_level[1] = [features.DarkOnesBlessing]
    features_by_level[6] = [features.DarkOnesOwnLuck]
    features_by_level[10] = [features.FiendishResilience]
    features_by_level[14] = [features.HurlThroughHell]


class GreatOldOne(SubClass):
    """Your patron is a mysterious entity whose nature is utterly foreign to the
    fabric of reality. It might come from the Far Realm, the space beyond
    reality, or it could be one of the elder gods known only in legends. Its
    motives are incomprehensible to mortals, and its knowledge so immense and
    ancient that even the greatest libraries pale in comparison to the vast
    secrets it holds. The Great Old One might be unaware of your existence or
    entirely indifferent to you, but the secrets you have learned allow you to
    draw your magic from it.

    Entities of this type include Ghaunadar, called That Which Lurks;
    Tharizdun, the Chained God; Dendar, the Night Serpent; Zargon, the
    Returner; Great Cthulhu; and other unfathomable beings.

    """
    name = "The Great Old One Patron"
    features_by_level = defaultdict(list)
    features_by_level[1] = [features.AwakenedMind]
    features_by_level[6] = [features.EntropicWard]
    features_by_level[10] = [features.ThoughtShield]
    features_by_level[14] = [features.CreateThrall]


# SCAG
class Undying(SubClass):
    """Death holds no sway over your patron, who has un- locked the secrets of
    everlasting life, although such a prize- like all power- comes at a
    price. Once mortal, the Undying has seen mortal lifetimes pass like the
    sea- sons, like the flicker of endless days and nights. It has the secrets
    of the ages to share, secrets of life and death. Beings of this sort
    include Vecna, Lord of the Hand and the Eye; the dread Iuz; the lich-queen
    Vol; the Undying Court of Aerenal; Vlaakith, lich-queen of the githyanki;
    and the deathless wizard Fistandantalus.

    In the Realms, Undying patrons include Larloch the Shadow King, legendary
    master of Warlock's Crypt, and Gilgeam, the God-King of Unther

    """
    name = "The Undying Patron"
    features_by_level = defaultdict(list)
    features_by_level[1] = [features.AmongTheDead]
    features_by_level[6] = [features.DefyDeath]
    features_by_level[10] = [features.UndyingNature]
    features_by_level[14] = [features.IndestructibleLife]


# XGTE
class Celestial(SubClass):
    """Your patron is a powerful being of the Upper Planes. You have bound yourself
    to an ancient empyrean, solar, ki-rin, unicorn, or other entity that
    resides in the planes of everlasting bliss. Your pact with that being
    allows you to experience the barest touch of the holy light that illu—
    minates the multiverse.

    Being connected to such power can cause changes in your behavior and
    beliefs. You might find yourself driven to annihilate the undead, to defeat
    fiends, and to protect the innocent. At times, your heart might also be
    filled with a longing for the celestial realm of your patron, and a desire
    to wander that paradise for the rest of your days. But you know that your
    mission is among mortals for now, and that your pact binds you to bring
    light to the dark places of the world.

    """
    name = "The Celestial Patron"
    spells_known = spells_prepared = (spells.Light, spells.SacredFlame)
    features_by_level = defaultdict(list)
    features_by_level[1] = [features.HealingLight]
    features_by_level[6] = [features.RadiantSoul]
    features_by_level[10] = [features.CelestialResilience]
    features_by_level[14] = [features.SearingVengeance]


class Hexblade(SubClass):
    """You have made your pact with a mysterious entity from the Shadowfell—a force
    that manifests in sentient magic weapons carved from the stuff of
    shadow. The mighty sword Blackrazor is the most notable of these weapons,
    which have been spread across the multiverse over the ages. The shadowy
    force behind these weapons can offer power to warlocks who form pacts with
    it. Many hexhlade warlocks create weapons that emulate those formed in the
    Shadowfell. Others forgo such arms, content to weave the dark magic of that
    plane into their spellcasting.

    Because the Raven Queen is known to have forged the first of these weapons,
    many sages speculate that she and the force are one and that the weapons,
    along with hexblade warlocks, are tools she uses to manipulate events on
    the Material Plane to her inscrutable ends

    """
    name = "Hexblade Patron"
    weapon_proficiencies = (weapons.MartialWeapon,)
    _proficiencies_text = ['martial weapons', 'medium armor', 'shields']
    features_by_level = defaultdict(list)
    features_by_level[1] = [features.HexbladesCurse, features.HexWarrior]
    features_by_level[6] = [features.AccursedSpecter]
    features_by_level[10] = [features.ArmorOfHexes]
    features_by_level[14] = [features.MasterOfHexes]


class Warlock(CharClass):
    name = 'Warlock'
    hit_dice_faces = 8
    subclass_select_level = 1
    saving_throw_proficiencies = ('wisdom', 'charisma')
    primary_abilities = ('charisma',)
    _proficiencies_text = ("light Armor", "simple weapons")
    class_skill_choices = ('Arcana', 'Deception', 'History',
                           'Intimidation', 'Investigation', 'Nature',
                           'Religion')
    weapon_proficiencies = (weapons.SimpleWeapon,)
    multiclass_weapon_proficiencies = weapon_proficiencies
    _multiclass_proficiencies_text = ('light armor', 'simple weapons')
    features_by_level = defaultdict(list)
    features_by_level[2] = [features.EldritchInvocation]
    features_by_level[3] = [features.PactBoon]
    features_by_level[11] = [features.MysticArcanum]
    features_by_level[20] = [features.EldritchMaster]
    subclasses_available = (Archfey, Fiend, GreatOldOne, Undying, Celestial,
                            Hexblade)
    spellcasting_ability = 'charisma'
    spell_slots_by_level = {
        1:  (2, 1, 0, 0, 0, 0, 0, 0, 0, 0),
        2:  (2, 2, 0, 0, 0, 0, 0, 0, 0, 0),
        3:  (2, 0, 2, 0, 0, 0, 0, 0, 0, 0),
        4:  (3, 0, 2, 0, 0, 0, 0, 0, 0, 0),
        5:  (3, 0, 0, 3, 0, 0, 0, 0, 0, 0),
        6:  (3, 0, 0, 3, 0, 0, 0, 0, 0, 0),
        7:  (3, 0, 0, 0, 2, 0, 0, 0, 0, 0),
        8:  (3, 0, 0, 0, 2, 0, 0, 0, 0, 0),
        9:  (3, 0, 0, 0, 0, 2, 0, 0, 0, 0),
        10: (4, 0, 0, 0, 0, 2, 0, 0, 0, 0),
        11: (4, 0, 0, 0, 0, 3, 0, 0, 0, 0),
        12: (4, 0, 0, 0, 0, 3, 0, 0, 0, 0),
        13: (4, 0, 0, 0, 0, 3, 0, 0, 0, 0),
        14: (4, 0, 0, 0, 0, 3, 0, 0, 0, 0),
        15: (4, 0, 0, 0, 0, 3, 0, 0, 0, 0),
        16: (4, 0, 0, 0, 0, 3, 0, 0, 0, 0),
        17: (4, 0, 0, 0, 0, 4, 0, 0, 0, 0),
        18: (4, 0, 0, 0, 0, 4, 0, 0, 0, 0),
        19: (4, 0, 0, 0, 0, 4, 0, 0, 0, 0),
        20: (4, 0, 0, 0, 0, 4, 0, 0, 0, 0),
    }



from .. import (weapons, features)
from .classes import CharClass, SubClass
from collections import defaultdict


# PHB
class Thief(SubClass):
    """You hone your skills in the larcenous arts. Burglars, bandits, cutpurses,
    and other criminals typically follow this archetype, but so do rogues who
    prefer to think of themselves as professional treasure seekers, explorers,
    delvers, and investigators. In addition to improving your agility and
    stealth, you learn skills useful for delving into ancient ruins, reading
    unfamiliar languages, and using magic items you normally couldn’t employ

    """
    name = "Thief"
    features_by_level = defaultdict(list)
    features_by_level[3] = [features.FastHands, features.SecondStoryWork]
    features_by_level[9] = [features.SupremeSneak]
    features_by_level[13] = [features.UseMagicDevice]
    features_by_level[17] = [features.ThiefsReflexes]


class Assassin(SubClass):
    """You focus your training on the grim art of death. Those who adhere to this
    archetype are diverse: hired killers, spies, bounty hunters, and even
    specially anointed priests trained to exterminate the enemies of their
    deity. Stealth, poison, and disguise help you eliminate your foes with
    deadly efficiency

    """
    name = "Assassin"
    _proficiencies_text = ('disguise kit', "poisoner's kit")
    features_by_level = defaultdict(list)
    features_by_level[3] = [features.Assassinate]
    features_by_level[9] = [features.InfiltrationExpertise]
    features_by_level[13] = [features.Imposter]
    features_by_level[17] = [features.DeathStrike]


class ArcaneTrickster(SubClass):
    """Some rogues enhance their fine-honed skills of stealth and agility with
    magic, learning tricks of enchantment and illusion. These rogues include
    pickpockets and burglars, but also pranksters, mischief-makers, and a
    significant number of adventurers.

    """
    name = "Arcane Trickster"
    features_by_level = defaultdict(list)
    features_by_level[3] = [features.MageHandLegerdemain]
    features_by_level[9] = [features.MagicalAmbush]
    features_by_level[13] = [features.VersatileTrickster]
    features_by_level[17] = [features.SpellThief]
    spellcasting_ability = 'intelligence'
    multiclass_spellslots_by_level = {
        # char_lvl: (cantrips, 1st, 2nd, 3rd, ...)
        1:  (0, 0, 0, 0, 0, 0, 0, 0, 0, 0),
        2:  (0, 0, 0, 0, 0, 0, 0, 0, 0, 0),
        3:  (3, 2, 0, 0, 0, 0, 0, 0, 0, 0),
        4:  (3, 3, 0, 0, 0, 0, 0, 0, 0, 0),
        5:  (3, 3, 0, 0, 0, 0, 0, 0, 0, 0),
        6:  (3, 3, 0, 0, 0, 0, 0, 0, 0, 0),
        7:  (3, 4, 2, 0, 0, 0, 0, 0, 0, 0),
        8:  (3, 4, 2, 0, 0, 0, 0, 0, 0, 0),
        9:  (3, 4, 2, 0, 0, 0, 0, 0, 0, 0),
        10: (4, 4, 3, 0, 0, 0, 0, 0, 0, 0),
        11: (4, 4, 3, 0, 0, 0, 0, 0, 0, 0),
        12: (4, 4, 3, 0, 0, 0, 0, 0, 0, 0),
        13: (4, 4, 3, 2, 0, 0, 0, 0, 0, 0),
        14: (4, 4, 3, 2, 0, 0, 0, 0, 0, 0),
        15: (4, 4, 3, 2, 0, 0, 0, 0, 0, 0),
        16: (4, 4, 3, 3, 0, 0, 0, 0, 0, 0),
        17: (4, 4, 3, 3, 0, 0, 0, 0, 0, 0),
        18: (4, 4, 3, 3, 0, 0, 0, 0, 0, 0),
        19: (4, 4, 3, 3, 1, 0, 0, 0, 0, 0),
        20: (4, 4, 3, 3, 1, 0, 0, 0, 0, 0),
    }


# XGTE
class Inquisitive(SubClass):
    """As an archetypal Inquisitive, you excel at rooting out se- crets and
    unraveling mysteries. You rely on your sharp eye for detail, but also on
    your finely honed ability to read the words and deeds of other creatures to
    deter- mine their true intent. You excel at defeating creatures that hide
    among and prey upon ordinary folk, and your mastery of lore and your keen
    deductions make you well equipped to expose and end hidden evils

    """
    name = "Inquisitive"
    features_by_level = defaultdict(list)
    features_by_level[3] = [features.EarForDeceit, features.EyeForDetail,
                            features.InsightfulFighting]
    features_by_level[9] = [features.SteadyEye]
    features_by_level[13] = [features.UnerringEye]
    features_by_level[17] = [features.EyeForWeakness]


class Mastermind(SubClass):
    """Your focus is on people and on the influence and secrets they have. Many
    spies, courtiers, and schemers follow this archetype, leading lives of
    intrigue. Words are your weapons as often as knives or poison, and secrets
    and favors are some of your favorite treasures.

    """
    name = "Mastermind"
    features_by_level = defaultdict(list)
    features_by_level[3] = [features.MasterOfIntrigue,
                            features.MasterOfTactics]
    features_by_level[9] = [features.InsightfulManipulator]
    features_by_level[13] = [features.Misdirection]
    features_by_level[17] = [features.SoulOfDeceit]


class Scout(SubClass):
    """You are skilled in stealth and surviving far from the streets of a city,
    allowing you to scout ahead of your companions during expeditions. Rogues
    who embrace this archetype are at home in the wilderness and among
    barbarians and rangers, and many Scouts serve as the eyes and ears of war
    bands. Ambusher, spy, bounty hunter#these are just a few of the roles that
    Scouts as- sume as they range the world.

    """
    name = "Scout"
    skill_proficiencies = ("nature", "survival")
    features_by_level = defaultdict(list)
    features_by_level[3] = [features.Skirmisher, features.Survivalist]
    features_by_level[9] = [features.SuperiorMobility]
    features_by_level[13] = [features.AmbushMaster]
    features_by_level[17] = [features.SuddenStrike]


class Swashbuckler(SubClass):
    """You focus your training on the art of the blade, relying on speed,
    elegance, and charm in equal parts. While some warriors are brutes clad in
    heavy armor, your method of fighting looks almost like a performance. Du—
    elists and pirates typically belong to this archetype. A Swashbuckler
    excels in single combat, and can fight with two weapons while safely
    darting away from an opponent

    """
    name = "Swashbuckler"
    features_by_level = defaultdict(list)
    features_by_level[3] = [features.FancyFootwork, features.RakishAudacity]
    features_by_level[9] = [features.Panache]
    features_by_level[13] = [features.ElegantManeuver]
    features_by_level[17] = [features.MasterDuelist]


class Rogue(CharClass):
    name = 'Rogue'
    hit_dice_faces = 8
    subclass_select_level = 3
    saving_throw_proficiencies = ('dexterity', 'intelligence')
    primary_abilities = ('dexterity',)
    _proficiencies_text = (
        'light armor', 'simple weapons', 'hand crossbows', 'longswords',
        'rapiers', 'shortswords', "thieves' tools")
    weapon_proficiencies = (weapons.SimpleWeapon, weapons.HandCrossbow,
                            weapons.Longsword, weapons.Rapier,
                            weapons.Shortsword)
    multiclass_weapon_proficiencies = ()
    _multiclass_proficiencies_text = ('light armor', "thieves' tools",
                                      '[choose one skill from Rogue list]')
    class_skill_choices = ('Acrobatics', 'Athletics', 'Deception',
                           'Insight', 'Intimidation', 'Investigation',
                           'Perception', 'Performance', 'Persuasion',
                           'Sleight of Hand', 'Stealth')
    num_skill_choices = 4
    features_by_level = defaultdict(list)
    features_by_level[1] = [features.RogueExpertise, features.SneakAttack]
    features_by_level[2] = [features.CunningAction]
    features_by_level[5] = [features.UncannyDodge]
    features_by_level[7] = [features.Evasion]
    features_by_level[11] = [features.ReliableTalent]
    features_by_level[14] = [features.BlindSense]
    features_by_level[15] = [features.SlipperyMind]
    features_by_level[18] = [features.Elusive]
    features_by_level[20] = [features.StrokeOfLuck]
    subclasses_available = (Thief, Assassin, ArcaneTrickster,
                            Inquisitive, Mastermind, Scout, Swashbuckler)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # slippery mind feature
        if self.owner.Rogue.level >= 15:
            self.saving_throw_proficiencies = ('dexterity', 'intelligence',
                                               'wisdom')



from .spells import Spell


class AbiDalzimsHorridWilting(Spell):
    """You draw the moisture from every creature in a 30-foot cube centered on a point 
    you choose within range. Each creature in that area must make a Constitution 
    saving throw. Constructs and undead aren’t affected, and plants and water 
    elementals make this saving throw with disadvantage. A creature takes 12d8 
    necrotic damage on a failed save, or half as much damage on a successful one. 
    
    Nonmagical plants in the area that aren’t creatures, such as trees and shrubs, 
    wither and die instantly.
    """
    name = "Abi-Dalzims Horrid Wilting"
    level = 8
    casting_time = "1 action"
    casting_range = "150 feet"
    components = ('V', 'S', 'M')
    materials = """A bit of sponge"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Necromancy"
    classes = ('Sorcerer', 'Wizard')


class AbsorbElements(Spell):
    """1 Reaction, which you take when you take acid, cold, fire, lightning, or thunder
     damage
    
    The spell captures some of the incoming energy, lessening its effect on
     you and storing it for your next melee attack. You have resistance to the 
    triggering damage type until the start of your next turn. Also, the first time 
    you hit with a melee attack on your next turn, the target takes an extra 1d6 
    damage of the triggering type, and the spell ends.
    
    At Higher Levels: When you 
    cast this spell using a spell slot of 2nd level or higher, the extra damage 
    increases by 1d6 for each slot level above 1st.
    """
    name = "Absorb Elements"
    level = 1
    casting_time = "Special"
    casting_range = "Self"
    components = ('S',)
    materials = """"""
    duration = "1 round"
    ritual = False
    magic_school = "Abjuration"
    classes = ('Druid', 'Ranger', 'Wizard', 'Sorcerer')


class AcidSplash(Spell):
    """You hurl a bubble of acid. 
    Choose one creature within range, or choose two 
    creatures within range that are within 5 feet of each other. A target must 
    succeed on a Dexterity saving throw or take 1d6 acid damage.
    
    At Higher Levels: 
    This spell’s damage increases by 1d6 when you reach 5th level (2d6), 11th level 
    (3d6), and 17th level (4d6).
    """
    name = "Acid Splash"
    level = 0
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ('V', 'S')
    materials = """"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Conjuration"
    classes = ('Sorcerer', 'Wizard')


class AganazzarsScorcher(Spell):
    """A line of roaring flame 30 feet long and 5 feet wide emanates from you in a 
    direction you choose. 
    Each creature in the line must make a Dexterity saving 
    throw. A creature takes 3d8 fire damage on a failed save, or half as much damage
     on a successful one.
    
    At Higher Levels: When you cast this spell using a spell 
    slot of 3rd level or higher, the damage increases by 1d8 for each slot level 
    above 2nd.
    """
    name = "Aganazzars Scorcher"
    level = 2
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ('V', 'S', 'M')
    materials = """A red dragon’s scale"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ('Sorcerer', 'Wizard')


class Aid(Spell):
    """Your spell bolsters your allies with toughness and resolve. 
    Choose up to three 
    creatures within range. Each target’s hit point maximum and current hit points 
    increase by 5 for the duration.
    
    At Higher Levels: When you cast this spell 
    using a spell slot of 3rd level or higher, a target’s hit points increase by an 
    additional 5 for each slot level above 2nd.
    """
    name = "Aid"
    level = 2
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ('V', 'S', 'M')
    materials = """A tiny strip of white cloth"""
    duration = "8 hours"
    ritual = False
    magic_school = "Abjuration"
    classes = ('Cleric', 'Paladin')


class Alarm(Spell):
    """You set an alarm against unwanted intrusion. 
    Choose a door, a window, or an 
    area within range that is no larger than a 20-foot cube. Until the spell ends, 
    an alarm alerts you whenever a tiny or larger creature touches or enters the 
    warded area. When you cast the spell, you can designate creatures that won’t set
     off the alarm. You also choose whether the alarm is mental or audible. 
    
    A 
    mental alarm alerts you with a ping in your mind if you are within 1 mile of the
     warded area. This ping awakens you if you are sleeping. 
    An audible alarm 
    produces the sound of a hand bell for 10 seconds within 60 feet.
    """
    name = "Alarm"
    level = 1
    casting_time = "1 minute"
    casting_range = "30 feet"
    components = ('V', 'S', 'M')
    materials = """A tiny bell and a piece of fine silver wire"""
    duration = "8 hours"
    ritual = True
    magic_school = "Abjuration"
    classes = ('Ranger', 'Wizard')


class AlterSelf(Spell):
    """You assume a different form.
    When you cast the spell, choose one of the 
    following options, the effects of which last for the duration of the spell. 
    While the spell lasts, you can end one option as an action to gain the benefits 
    of a different one.
    
    Aquatic Adaptation. You adapt your body to an aquatic 
    environment, sprouting gills, and growing webbing between your fingers. You can 
    breathe underwater and gain a swimming speed equal to your walking speed.
    Change
     Appearance. You transform your appearance. You decide what you look like, 
    including your height, weight, facial features, sound of your voice, hair 
    length, coloration, and distinguishing characteristics, if any. You can make 
    yourself appear as a member of another race, though none of your statistics 
    change. You also don’t appear as a creature of a different size than you, and 
    your basic shape stays the same; if you're bipedal, you can’t use this spell to 
    become quadrupedal, for instance. At any time for the duration of the spell, you
     can use your action to change your appearance in this way again.
    Natural 
    Weapons. You grow claws, fangs, spines, horns, or a different natural weapon of 
    your choice. Your unarmed strikes deal 1d6 bludgeoning, piercing, or slashing 
    damage, as appropriate to the natural weapon you chose, and you are proficient 
    with your unarmed strikes. Finally, the natural weapon is magic and you have a 
    +1 bonus to the attack and damage rolls you make using it.
    """
    name = "Alter Self"
    level = 2
    casting_time = "1 action"
    casting_range = "Self"
    components = ('V', 'S')
    materials = """"""
    duration = "Concentration, up to 1 hour"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Sorcerer', 'Wizard')


class AnimalFriendship(Spell):
    """This spell lets you convince a beast that you mean it no harm.
    Choose a beast 
    that you can see within range. It must see and hear you. If the beast’s 
    Intelligence is 4 or higher, the spell fails. Otherwise, the beast must succeed 
    on a Wisdom saving throw or be charmed by you for the spell’s duration. If you 
    or one of your companions harms the target, the spell ends.
    
    At Higher Levels: 
    When you cast this spell using a spell slot of 2nd level or higher, you can 
    affect one additional beast for each slot level above 1st.
    """
    name = "Animal Friendship"
    level = 1
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ('V', 'S', 'M')
    materials = """A morsel of food"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Enchantment"
    classes = ('Bard', 'Druid', 'Ranger')


class AnimalMessenger(Spell):
    """By means of this spell, you use an animal to deliver a message.
    Choose a Tiny 
    beast you can see within range, such as a squirrel, a blue ray, or a bat. You 
    specify a location, which you must have visited, and a recipient who matches a 
    general description, such as a man or woman dressed in the uniform of the town 
    guard or a red-haired dwarf wearing a pointed hat. You also speak a message of 
    up to twenty-five words. The target beast travels for the duration of the spell 
    towards the specified location, covering about 50 miles per 24 hours for a 
    flying messenger or 25 miles for other animals.
    
    When the messenger arrives, it 
    delivers your message to the creature that you described, replicating the sound 
    of your voice. The messenger speaks only to a creature matching the description 
    you gave. If the messenger doesn’t reach its destination before the spell ends, 
    the message is lost, and the beast makes it way back to where you cast this 
    spell.
    
    At Higher Levels: If you cast this spell using a spell slot of 3rd level
     or higher, the duration of the spell increases by 48 hours for each slot level 
    above 2nd.
    """
    name = "Animal Messenger"
    level = 2
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ('V', 'S', 'M')
    materials = """A morsel of food"""
    duration = "24 hours"
    ritual = True
    magic_school = "Enchantment"
    classes = ('Bard', 'Druid', 'Ranger')


class AnimalShapes(Spell):
    """Your magic turns others into beasts. 
    Choose any number of willing creatures 
    that you can see within range. You transform each target into the form of a 
    large or smaller beast with a challenge rating of 4 or lower. On subsequent 
    turns, you can use your actions to transform affected creatures into new forms. 
    
    
    The transformation lasts for the duration for each target, or until the target
     drops to 0 hit points or dies. You can choose a different form for each target.
     A target’s game statistics are replaced by the statistics of the chosen beast, 
    though the target retains its alignment and Intelligence, Wisdom, and Charisma 
    scores. The target assumes the hit points of its new form, and when it reverts 
    to its normal form, it returns to the number of hit point it had before it 
    transformed. If it reverts as a result of dropping to 0 hit points, any excess 
    damage carries over to its normal form. As long as the excess damage doesn’t 
    reduce the creature’s normal form to 0 hit points, it isn’t knocked unconcious. 
    The creature is limited in the actions it can perform by the nature of its new 
    form, and it can’t speak or cast spells. 
     
    The target’s gear melds into the new
     form. The target can’t activate, wield, or otherwise benefit from any of its 
    equipment.
    """
    name = "Animal Shapes"
    level = 8
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ('V', 'S')
    materials = """"""
    duration = "Concentration, up to 24 hours"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Druid',)


class AnimateDead(Spell):
    """This spell creates an undead servant. 
    Choose a pile of bones or a corpse of a 
    Medium or Small humanoid within range. Your spell imbues the target with a foul 
    mimicry of life, raising it as an undead creature. The target becomes a skeleton
     if you chose bones or a zombie if you chose a corpse (the DM has the creature’s
     game statistics). 
    
    On each of your turns, you can use a bonus action to 
    mentally command any creature you made with this spell if the creature is within
     60 feet of you (if you control multiple creatures, you can command any or all 
    of them at the same time, issuing the same command to each one). You decide what
     action the creature will take and where it will move during its next turn, or 
    you can issue a general command, such as to guard a particular chamber or 
    corridor. If you issue no commands, the creature only defends itself against 
    hostile creatures. Once given an order, the creature continues to follow it 
    until its task is complete.
    
    The creature is under your control for 24 hours, 
    after which it stops obeying any command you’ve given it. To maintain the 
    control of the creature for another 24 hours, you must cast this spell on the 
    creature again before the current 24-hour period ends. This use of the spell 
    reasserts your control over up to four creatures you have animated with this 
    spell, rather than animating a new one.
    
    At Higher Levels: When you cast this 
    spell using a spell slot of 4th level or higher, you animate or reassert control
     over two additional undead creatures for each slot level above 3rd. Each of the
     creatures must come from a different corpse or pile of bones.
    """
    name = "Animate Dead"
    level = 3
    casting_time = "1 minute"
    casting_range = "10 feet"
    components = ('V', 'S', 'M')
    materials = """A drop of blood, a piece of flesh, and a pinch of bone dust"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Necromancy"
    classes = ('Cleric', 'Wizard')


class AnimateObjects(Spell):
    """Objects come to life at your command. 
    Choose up to ten nonmagical objects 
    within range that are not being worn or carried. Medium targets count as two 
    objects, Large targets count as four objects, Huge targets count as eight 
    objects. You can’t animate any object larger than Huge. Each target animates and
     becomes a creature under your control until the spell ends or until reduced to 
    0 hit points. 
    
    As a bonus action, you can mentally command any creature you 
    made with this spell if the creature is within 500 feet of you (if you control 
    multiple creatures, you can command any or all of them at the same time, issuing
     the same command to each one). You decide what action the creature will take 
    and where it will move during its next turn, or you can issue a general command,
     such as to guard a particular chamber or corridor. If you issue no commands, 
    the creature only defends itself against hostile creatures. Once given an order,
     the creature continues to follow it until its task is complete. 
    
    Animated 
    Object Statistics 
    Tiny – HP: 20, AC: 18, Attack: +8 to hit, 1d4 + 4 damage, 
    Str: 4, Dex: 18 
    Small – HP: 25, AC: 16, Attack: +6 to hit, 1d8 + 2 damage, Str:
     6, Dex: 14 
    Medium – HP: 40, AC: 13, Attack: +5 to hit, 2d6 + 1 damage, Str: 
    10, Dex: 12 
    Large – HP: 50, AC: 10, Attack: +6 to hit, 2d10 + 2 damage, Str: 
    14, Dex: 10 
    Huge – HP: 80, AC: 10, Attack: +8 to hit, 2d12 + 4 damage, Str: 18,
     Dex: 6 
    
    An animated object is a construct with AC, hit points, attacks, 
    Strength, and Dexterity determine by its size. Its Constitution is 10 and its 
    Intelligence and Wisdom are 3, and its Charisma is 1. Its speed is 30 feet; if 
    the objects lack legs or other appendages it can use for locomotion, it instead 
    has a flying speed of 30 feet and can hover. If the object is securely attached 
    to a surface or larger object, such as a chain bolted to a wall, its speed is 0.
     It has blindsight with a radius of 30 feet and is blind beyond that distance. 
    When the animated object drops to 0 hit points, it reverts to its original 
    object form, and any remaining damage carries over to its original object form.
     
    
    If you command an object to attack, it can make a single melee attack against
     a creature within 5 feet of it. It makes a slam attack with an attack bonus and
     bludgeoning damage determine by its size. The DM might rule that a specific 
    object inflicts slashing or piercing damage based on its form.
    
    At Higher 
    Levels: If you cast this spell using a spell slot of 6th level or higher, you 
    can animate two additional objects for each slot level above 5th.
    """
    name = "Animate Objects"
    level = 5
    casting_time = "1 action"
    casting_range = "120 feet"
    components = ('V', 'S')
    materials = """"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Bard', 'Sorcerer', 'Wizard')


class AntilifeShell(Spell):
    """A shimmering barrier extends out from you in a 10-foot radius and moves with 
    you, remaining centered on you and hedging out creatures other than undead and 
    constructs. 
    The barrier lasts for the duration. The barrier prevents an 
    affected creature from passing or reaching through. An affected creature can 
    cast spells or make attacks with ranged or reach weapons through the barrier. 
    
    If you move so that an affect creature is forced to pass through the barrier, 
    the spell ends.
    """
    name = "Antilife Shell"
    level = 5
    casting_time = "1 action"
    casting_range = "Self (10-foot radius)"
    components = ('V', 'S')
    materials = """"""
    duration = "Concentration, up to 1 hour"
    ritual = False
    magic_school = "Abjuration"
    classes = ('Druid',)


class AntimagicField(Spell):
    """A 10-foot-radius invisible sphere of antimagic surrounds you. 
    This area is 
    divorced from the magical energy that suffeses the multiverse. Within the 
    sphere, spells can’t be cast, summoned creatures disappear, and even magic items
     become mundane. Until the spell ends, the spere moves with you, centered on 
    you. 
    Spells and other magical effects, except those created by an artifact or a
     deity, are suppressed in the sphere and can’t protrude into it. A slot expended
     to cast a suppressed spell is consumed. While an effect is suppressed, it 
    doesn’t function, but the time it spends suppressed counts against its duration.
     
    
    Targeted Effects. 
    Spells and other magical effects, such as magic missle and
     charm person, that target a creature or an object in the sphere have no effect 
    on that target. 
    
    Areas of Magic. 
    The area of another spell or magical effect, 
    such as fireball, can’t extend into the sphere. If the sphere overlaps an area 
    of magic, the part of the area that is covered by the sphere is suppressed. For 
    example, the flames created by a wall of fire are suppressed within the sphere, 
    creating a gap in the wall if the overlap is large enough. 
    
    Spells. 
     Any 
    active spell or other magical effect on a creature or an object in the sphere is
     suppressed while the creature or object is in it. 
    
    Magic Items. 
    The 
    properties and powers of magic items are suppressed in the sphere. Forexample, a
     +1 longsword in the sphere functions as a nonmagical longsword. A magic 
    weapon’s properties and powers are suppressed if it is used against a target in 
    the sphere or wielded by an attacker in the sphere. If a magic weapon or piece 
    of magic ammunition fully leaves the sphere (For example, if you fire a magic 
    arrow or throw a magic spear at a target outside the sphere), the magic of the 
    item ceases to be supressed as soon as it exits. 
    
    Magical Travel. 
    
    Teleportation and planar travel fail to work in the sphere, whether the sphere 
    is the destination or the departure point for such magical travel. A portal to 
    another location, world, or plane of existence, as well as an opening to an 
    extradimensional space such as that created by the rope trick spells, 
    temporarily closes while in the sphere. 
    
    Creatures and Objects. 
    A creature or 
    object summoned or created by magic temporarily winks out of existence in the 
    sphere. Such a creature instantly reappears once the space the creature occupied
     is no longer withinthe sphere. 
    
    Dispel Magic. 
    Spells and magical effects such
     as dispel magic have no effect on the sphere. Likewise, the spheres created by 
    different antimagic field spells don’t nullify each other.
    """
    name = "Antimagic Field"
    level = 8
    casting_time = "1 action"
    casting_range = "Self (10-foot-radius sphere)"
    components = ('V', 'SM')
    materials = """A pinch of powdered iron or iron filings"""
    duration = "Concentration, up to 1 hour"
    ritual = False
    magic_school = "Abjuration"
    classes = ('Cleric', 'Wizard')


class Antipathysympathy(Spell):
    """This spell attracts or repels creatures of your choice. 
    You target something 
    within range, either a Huge or smaller object or creature or an area that is no 
    larger than a 200-foot cube. Then specify a kind of intelligent creature, such 
    as red dragons, goblins, or vampires. You invest the target with an aura that 
    either attracts or repels the specified creatures for the duration. Choose 
    antipathy or sympathy as the aura’s effect. 
    
    Antipathy. 
    The enchantment causes
     creatures of the kind you designated to feel an intense urge to leave the area 
    and avoid the target. When such a creature can see the target or comes within 60
     feet of it, the creature must succeed on a Wisdom saving throw or 
    become frightened. The creature remains frightened while it can see the target 
    or is within 60 feet of it. While frightened by the target, the creature must 
    use its movement to move to the nearest safe spot from which it can’t see the 
    target. If the creature moves more than 60 feet from the target and can’t see 
    it, the creature is no longer frightened, but the creature becomes frightened 
    again if it regains sight of the target or moves within 60 feet of it. 
    
    
    Sympathy. 
    The enchantment causes the specified creatures to feel an intense 
    urge to approach the target while within 60 feet of it or able to see it. When 
    such a creature can see the target or comes within 60 feet o fit, the creature 
    must succeed on a Wisdom saving throw or use its movement on each of its turns 
    to enter the area or move within reach of the target. When the creature has done
     so, it can’t willingly move away from the target. If the target damages or 
    otherwise harms an affected creature, the affected creature can make a 
    Wisdom saving throw to end the effect, as described below. 
    
    Ending the Effect.
     
    If an affected creature ends its turn while not within 60 feet of the target 
    or able to see it, the creature makes a Wisdom saving throw. On a successful 
    save, the creature is no longer affected by the target and recognizes the 
    feeling of repugnance or attraction as magical. In addition, a creature affected
     by the spell is allowed another Wisdom saving throw every 24 hours while the 
    spell persists. 
    A creature that successfully saves against this effect is 
    immune to it for 1 minute, after which time it can be affected again.
    """
    name = "Antipathysympathy"
    level = 8
    casting_time = "1 hour"
    casting_range = "60 feet"
    components = ('V', 'S', 'M')
    materials = """Either a lump of alum soaked in vinegar for the antipathy effect or a drop of honey for the sympathy effect"""
    duration = "10 days"
    ritual = False
    magic_school = "Enchantment"
    classes = ('Druid', 'Wizard')


class ArcaneEye(Spell):
    """You create an invisible, magical eye within range that hovers in the air for the
     duration. 
    You mentally receive visual information from the eye, which has 
    normal vision and darkvision out to 30 feet. The eye can look in every 
    direction. 
    As an action, you can move the eye up to 30 feet in any direction. 
    There is no limit to how far away from you the eye can move, but it can’t enter 
    another plane of existence. A solid barrier blocks the eye’s movement, but the 
    eye can pass through an opening as small as 1 inch in diameter.
    """
    name = "Arcane Eye"
    level = 4
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ('V', 'S', 'M')
    materials = """A bit of bat fur"""
    duration = "Concentration, up to 1 hour"
    ritual = False
    magic_school = "Divination"
    classes = ('Wizard',)


class ArcaneGate(Spell):
    """You create linked teleportation portals that remain open for the duration. 
    
    Choose two points on the ground that you can see, one point within 10 feet of 
    you and one point within 500 feet of you. A circular portal, 10 feet in 
    diameter, opens over each point. If the portal would open in the space occupied 
    by a creature, the spell fails, and the casting is lost. 
    
     The portals are two-
    dimensional glowing rings filled with mist, hovering inches from the ground and 
    perpendicular to it at the points you choose. A ring is visible only from one 
    side (your choice), which is the side that functions as a portal. 
     
    Any 
    creature or object entering the portal exits from the other portal as if the two
     were adjacent to each other; passing through a portal from the nonportal side 
    has no effect. The mist that fills each portal is opaque and blocks vision 
    through it. On your turn, you can rotate the rings as a bonus action so that the
     active side faces in a different direction.
    """
    name = "Arcane Gate"
    level = 6
    casting_time = "1 action"
    casting_range = "500 feet"
    components = ('V', 'S')
    materials = """"""
    duration = "Concentration, up to 10 minutes"
    ritual = False
    magic_school = "Conjuration"
    classes = ('Sorcerer', 'Warlock', 'Wizard')


class ArcaneLock(Spell):
    """You touch a closed door, window, gate, chest, or other entryway, and 
    it becomes locked for the duration. You and the creatures you 
    designate when you cast this spell can open the object normally. You 
    can also set a password that, when spoken within 5 feet of the object,
     suppresses this spell for 1 minute. Otherwise, it is impassable until
     it is broken or the spell is dispelled or suppressed. Casting knock 
    on the object suppresses arcane lock for 10 minutes.
    
    While affected 
    by this spell, the object is more difficult to break or force open; 
    the DC to break it or pick any locks on it increases by 10.
    """
    name = "Arcane Lock"
    level = 2
    casting_time = "1 action"
    casting_range = "Touch"
    components = ('V', 'S', 'M')
    materials = """Gold dust worth at least 25 gp, which the spell consumes"""
    duration = "Until dispelled"
    ritual = False
    magic_school = "Abjuration"
    classes = ('Wizard',)


class ArmorOfAgathys(Spell):
    """A protective magical force surrounds you, manifesting as a spectral frost that 
    covers you and your gear. 
    You gain 5 temporary hit points for the duration. If 
    a creature hits you with a melee attack while you have these hit points, the 
    creature takes 5 cold damage.
    
    At Higher Levels: When you cast this spell using 
    a spell slot of 2nd level or higher, both the temporary hit points and the cold 
    damage increase by 5 for each slot
    """
    name = "Armor Of Agathys"
    level = 1
    casting_time = "1 action"
    casting_range = "Self"
    components = ('V', 'S', 'M')
    materials = """A cup of water"""
    duration = "1 hour"
    ritual = False
    magic_school = "Abjuration"
    classes = ('Warlock',)


class ArmsOfHadar(Spell):
    """You invoke the power of Hadar, the Dark Hunger. 
    Tendrils of dark energy erupt 
    from you and batter all creatures within 10 feet of you. Each creature in that 
    area must make a Strength saving throw. On a failed save, a target takes 2d6 
    necrotic damage and can’t take reactions until its next turn. On a successful 
    save, the creature takes half damage, but suffers no other effect.
    
    At Higher 
    Levels: When you cast this spell using a spell slot of 2nd level or higher, the 
    damage increases by 1d6 for each slot level above 1st.
    """
    name = "Arms Of Hadar"
    level = 1
    casting_time = "1 action"
    casting_range = "Self (10-foot radius)"
    components = ('V', 'S')
    materials = """"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Conjuration"
    classes = ('Warlock',)


class AstralProjection(Spell):
    """You and up to eight willing creatures within range project your astral bodies 
    into the Astral Plane (the spell fails and the casting is wasted if you are 
    already on that plane). 
    The material body you leave behind is unconscious and 
    in a state of suspended animation; it doesn’t need food or air and doesn’t age. 
    
    
    Your astral body resembles your mortal form in almost every way, replicating 
    your game statistics and possessions. The principal difference is the addition 
    of a silvery cord that extends from between your shoulder blades and trails 
    behind you, fading to invisibility after 1 foot. This cord is your tether to 
    your material body. As long as the tether remains intact, you can find your 
    way home. If the cord is cut something that can happen only when an effect 
    specifically states that it does your soul and body are separated, killing you 
    instantly. 
    
    Your astral form can freely travel through the Astral Plane and can
     pass through portals there leading to any other plane. If you enter a new plane
     or return to the plane you were on when casting this spell, your body 
    and possessions are transported along the silver cord, allowing you to re-enter 
    your body as you enter the new plane. Your astral form is a separate 
    incarnation. Any damage or other effects that apply to it have no effect on 
    your physical body, nor do they persist when you return to it. The spell ends 
    for you and your companions when you use your action to dismiss it. When the 
    spell ends, the affected creature returns to its physical body, and it awakens. 
    
    
    The spell might also end early for you or one of your companions. A successful
     dispel magic spell used against an astral or physical body ends the spell for 
    that creature. If a creature’s original body or its astral form drops to 0 hit 
    points, the spell ends for that creature. If the spell ends and the silver cord 
    is intact, the cord pulls the creature’s astral form back to its body, ending 
    its state of suspended animation. If you are returned to your body prematurely, 
    your companions remain in their astral forms and must find their own way back to
     their bodies, usually by dropping to 0 hit points.
    """
    name = "Astral Projection"
    level = 9
    casting_time = "1 hour"
    casting_range = "10 feet"
    components = ('V', 'S', 'M')
    materials = """For each creature you affect with this spell, you must provide one jacinth worth at least 1,000 gp and one ornately carved bar of silver worth at least 100 gp, all of which the spell consumes"""
    duration = "Special"
    ritual = False
    magic_school = "Necromancy"
    classes = ('Cleric', 'Warlock', 'Wizard')


class Augury(Spell):
    """By casting gem-inlaid sticks, rolling dragon bones, laying out ornate cards, or 
    employing some other divining tool, you receive an omen from an otherworldly 
    entity about the results of a specific course of action that you plan to take 
    within the next 30 minutes. The DM chooses from the following possible omens: 
    
    
    • Weal, for good results 
    • Woe, for bad results 
    • Weal and woe, for both good 
    and bad results 
    • Nothing, for results that aren’t especially good or bad 
    
    The
     spell doesn’t take into account any possible circumstances that might change 
    the outcome, such as the casting of additional spells or the loss or gain of a 
    companion.    If you cast the spell two or more times before completing your 
    next long rest, there is a cumulative 25 percent chance for each casting after 
    the first that you get a random reading. The DM makes this roll in secret.
    """
    name = "Augury"
    level = 2
    casting_time = "1 minute"
    casting_range = "Self"
    components = ('V', 'S', 'M')
    materials = """Specially marked sticks, bones, or similar tokens worth at least 25 gp"""
    duration = "Instantaneous"
    ritual = True
    magic_school = "Divination"
    classes = ('Cleric',)


class AuraOfLife(Spell):
    """Life-preserving energy radiates from you in an aura with a 30-foot radius. 
    
    Until the spll ends, the aura moves with you, centered on you. Each nonhostile 
    creature in the aura (including you) has resistance to necrotic damage, and its 
    hit point maximum can’t be reduced. In addition, a nonhostile, living creature 
    regains 1 hit point when it starts its turn in the arua with 0 hit points.
    """
    name = "Aura Of Life"
    level = 4
    casting_time = "1 action"
    casting_range = "Self (30-foot radius)"
    components = ('V',)
    materials = """"""
    duration = "Concentration, up to 10 minutes"
    ritual = False
    magic_school = "Abjuration"
    classes = ('Paladin',)


class AuraOfPurity(Spell):
    """Purifying energy radiates from you in an aura with a 30-foot radius. 
    Until the 
    spell ends, the aura moves with you, centered on you. Each nonhostile creature 
    in the aura (including you) can’t become diseased, has resistance to poison 
    damage, and has advantage on saving throws against effects that cause any of the
     following conditions: blnded, charmed, deafended, frightened, paralyzed, 
    poisoned, and stunned.
    """
    name = "Aura Of Purity"
    level = 4
    casting_time = "1 action"
    casting_range = "Self (30-foot radius)"
    components = ('V',)
    materials = """"""
    duration = "Concentration, up to 10 minutes"
    ritual = False
    magic_school = "Abjuration"
    classes = ('Paladin',)


class AuraOfVitality(Spell):
    """Healing energy radiates from you in an aura with a 30-foot radius. 
    Until the 
    spell ends, the aura moves with you, centered on you. You can use a bonus action
     to cause one creature in the aura (including you) to regain 2d6 hit points.
    """
    name = "Aura Of Vitality"
    level = 3
    casting_time = "1 action"
    casting_range = "Self (30-foot radius)"
    components = ('V',)
    materials = """"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Evocation"
    classes = ('Paladin',)


class Awaken(Spell):
    """After spending the casting time tracing magical pathways within a precious 
    gemstone, you touch a huge or smaller beast or plant. 
    The target must have 
    either no Intelligence score or an Intelligence of 3 or less. The target gains 
    an Intelligence of 10. The target also gains the ability to speak one language 
    you know. If the target is a plant, it gains the ability to move its limbs, 
    roots, vinces, creepers, and so forth, and it gains senses similar to a huamn’s.
     Your DM chooses statistics appropriate for the awakened plant, such as the 
    statistics for the awakened shrub or the awakened tree. 
    
    The awakened beast or 
    plant is charmed by you for 30 days or until you and your companions do anything
     harmful to it. When the charmed condition ends, the awakened creature chooses 
    whether to remain friendly to you, based on how you treated it while it was 
    charmed.
    """
    name = "Awaken"
    level = 5
    casting_time = "8 hours"
    casting_range = "Touch"
    components = ('V', 'S', 'M')
    materials = """An agate worth at least 1,000 gp, which the spell consumes"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Bard', 'Druid')





from .. import (weapons, features)
from .classes import CharClass, SubClass
from collections import defaultdict


# PHB
class Abjuration(SubClass):
    """The School of Abjuration emphasizes magic that blocks, banishes, or
    protects. Detractors of this school say that its tradition is about denial,
    negation rather than positive assertion. You understand, however, that
    ending harmful effects, protecting the weak, and banishing evil influences
    is anything but a philosophical void. It is a proud and respected vocation.

    Called abjurers, members of this school are sought when baleful spirits
    require exorcism, when important locations must be guarded against magical
    spying, and when portals to other planes of existence must be closed.

    """
    name = "School of Abjuration"
    features_by_level = defaultdict(list)
    features_by_level[2] = [features.AbjurationSavant, features.ArcaneWard]
    features_by_level[6] = [features.ProjectedWard]
    features_by_level[10] = [features.ImprovedAbjuration]
    features_by_level[14] = [features.SpellResistance]


class Conjuration(SubClass):
    """As a conjurer, you favor spells that produce objects and creatures out o f
    thin air. You can conjure billowing clouds of killing fog or summon
    creatures from elsewhere to fight on your behalf. As your mastery grows,
    you learn spells of transportation and can teleport yourself across vast
    distances, even to other planes of existence, in an instant

    """
    name = "School of Conjuration"
    features_by_level = defaultdict(list)
    features_by_level[2] = [features.ConjurationSavant, features.MinorIllusion]
    features_by_level[6] = [features.BenignTransposition]
    features_by_level[10] = [features.FocusedConjuration]
    features_by_level[14] = [features.DurableSummons]


class Divination(SubClass):
    """The counsel of a diviner is sought by royalty and commoners alike, for all
    seek a clearer understanding of the past, present, and future. As a
    diviner, you strive to part the veils of space, time, and consciousness so
    that you can see clearly. You work to master spells of discernment, remote
    viewing, supernatural knowledge, and foresight.

    """
    name = "School of Divination"
    features_by_level = defaultdict(list)
    features_by_level[2] = [features.DivinationSavant, features.Portent]
    features_by_level[6] = [features.ExpertDivination]
    features_by_level[10] = [features.TheThirdEye]
    features_by_level[14] = [features.GreaterPortent]


class Enchantment(SubClass):
    """As a member of the School of Enchantment, you have honed your ability to
    magically entrance and beguile other people and monsters. Some enchanters
    are peacemakers who bewitch the violent to lay down their arms and charm
    the cruel into showing mercy. Others are tyrants who magically bind the
    unwilling into their service. Most enchanters fall somewhere in between.

    """
    name = "School of Enchantment"
    features_by_level = defaultdict(list)
    features_by_level[2] = [features.EnchantmentSavant, features.HypnoticGaze]
    features_by_level[6] = [features.InstinctiveGaze]
    features_by_level[10] = [features.SplitEnchantment]
    features_by_level[14] = [features.AlterMemories]


class Evocation(SubClass):
    """You focus your study on magic that creates powerful elemental effects such
    as bitter cold, searing flame, rolling thunder, crackling lightning, and
    burning acid. Some evokers find employment in military forces, serving as
    artillery to blast enemy armies from afar. Others use their spectacular
    power to protect the weak, while some seek their own gain as bandits,
    adventurers, or aspiring tyrants.

    """
    name = "School of Evocation"
    features_by_level = defaultdict(list)
    features_by_level[2] = [features.EvocationSavant, features.SculptSpells]
    features_by_level[6] = [features.PotentCantrip]
    features_by_level[10] = [features.EmpoweredEvocation]
    features_by_level[14] = [features.Overchannel]


class Illusion(SubClass):
    """You focus your studies on magic that dazzles the senses, befuddles the
    mind, and tricks even the wisest folk. Your magic is subtle, but the
    illusions crafted by your keen mind make the impossible seem real. Some
    illusionists—including many gnome w izards—are benign tricksters who use
    their spells to entertain. Others are more sinister masters of deception,
    using their illusions to frighten and fool others for their personal gain.

    """
    name = "School of Illusion"
    features_by_level = defaultdict(list)
    features_by_level[2] = [features.IllusionSavant,
                            features.ImprovedMinorIllusion]
    features_by_level[6] = [features.MalleableIllusions]
    features_by_level[10] = [features.IllusorySelf]
    features_by_level[14] = [features.IllusoryReality]


class Necromancy(SubClass):
    """The School of Necromancy explores the cosm ic forces of life, death, and
    undeath. As you focus your studies in this tradition, you learn to
    manipulate the energy that animates all living things. As you progress, you
    learn to sap the life force from a creature as your magic destroys its
    body, transforming that vital energy into magical power you can manipulate.

    Most people see necromancers as menacing, or even villainous, due to the
    close association with death. Not all necromancers are evil, but the forces
    they manipulate are considered taboo by many societies

    """
    name = "School of Necromancy"
    features_by_level = defaultdict(list)
    features_by_level[2] = [features.NecromancySavant, features.GrimHarvest]
    features_by_level[6] = [features.UndeadThralls]
    features_by_level[10] = [features.InuredToUndeath]
    features_by_level[14] = [features.CommandUndead]


class Transmutation(SubClass):
    """You are a student of spells that modify energy and matter. To you, the
    world is not a fixed thing, but eminently mutable, and you delight in being
    an agent of change. You wield the raw stuff of creation and learn to alter
    both physical forms and mental qualities. Your magic gives you the tools to
    become a smith on reality’s forge.

    Some transmuters are tinkerers and pranksters, turning people into toads
    and transforming copper into silver for fun and occasional profit. Others
    pursue their magical studies with deadly seriousness, seeking the power of
    the gods to make and destroy worlds

    """
    name = "School of Transmutation"
    features_by_level = defaultdict(list)
    features_by_level[2] = [features.TransmutationSavant,
                            features.MinorAlchemy]
    features_by_level[6] = [features.TransmutersStone]
    features_by_level[10] = [features.Shapechanger]
    features_by_level[14] = [features.MasterTransmuter]


# SCAG
class Bladesinging(SubClass):
    """**Restriction: Elves Only**
    
    Bladesingers are elves who bravely defend their people and lands. They are
    elf wizards who master a school of sword fighting grounded in a tradition
    of arcane magic. In combat, a bladesinger uses a series of intricate,
    elegant maneuvers that fend off harm and allow the bladesinger to channel
    magic into devastating attacks and a cunning defense

    """
    name = "School of Bladesinging"
    _proficiencies_text = ('light armor',)
    skill_proficiencies = ('performance',)
    features_by_level = defaultdict(list)
    features_by_level[2] = [features.Bladesong,]
    features_by_level[6] = [features.ExtraAttackBladesinging]
    features_by_level[10] = [features.SongOfDefense]
    features_by_level[14] = [features.SongOfVictory]


# XGTE
class WarMagic(SubClass):
    """A variety of arcane colleges specialize in training wiz— ards for war. The
    tradition of War Magic blends principles of evocation and abjuration,
    rather than specializing in either of those schools. It teaches
    techniques that empower a caster’s spells, while also providing methods for
    wizards to bolster their own defenses.

    Followers of this tradition are known as war mages.  They see their magic
    as both a weapon and armor, a resource superior to any piece of steel. War
    mages act fast in battle, using their spells to seize tactical control of a
    situation. Their spells strike hard, while their defensive skills foil
    their opponents“ attempts to counterattack. War mages are also adept at
    turning other spellcasters’ magical energy against them.

    In great battles, a war mage often works with evokers, abjurers, and other
    types of wizards. Evokers, in particular, sometimes tease war mages for
    splitting their attention between offense and defense. A war mage’s typical
    response: “What good is being able to throw a mighty fireball if I die
    before I can cast it?

    """
    name = "School of War Magic"
    features_by_level = defaultdict(list)
    features_by_level[2] = [features.ArcaneDeflection, features.TacticalWit]
    features_by_level[6] = [features.PowerSurge]
    features_by_level[10] = [features.DurableMagic]
    features_by_level[14] = [features.DeflectingShroud]


class Wizard(CharClass):
    name = 'Wizard'
    hit_dice_faces = 6
    subclass_select_level = 2
    saving_throw_proficiencies = ('intelligence', 'wisdom')
    primary_abilities = ('intelligence',)
    _proficiencies_text = ('daggers', 'darts', 'slings',
                           'quarterstaffs', 'light crossbows')
    weapon_proficiencies = (weapons.Dagger, weapons.Dart,
                            weapons.Sling, weapons.Quarterstaff,
                            weapons.LightCrossbow)
    multiclass_weapon_proficiencies = ()
    _multiclass_proficiencies_text = ()
    class_skill_choices = ('Arcana', 'History', 'Investigation',
                           'Medicine', 'Religion')
    features_by_level = defaultdict(list)
    features_by_level[1] = [features.ArcaneRecovery]
    features_by_level[18] = [features.SpellMastery]
    features_by_level[18] = [features.SignatureSpells]
    subclasses_available = (Abjuration, Conjuration, Divination, Enchantment,
                            Evocation, Illusion, Necromancy, Transmutation,
                            Bladesinging, WarMagic)
    spellcasting_ability = 'intelligence'
    spell_slots_by_level = {
        # char_lvl: (cantrips, 1st, 2nd, 3rd, ...)
        1:  (3, 2, 0, 0, 0, 0, 0, 0, 0, 0),
        2:  (3, 3, 0, 0, 0, 0, 0, 0, 0, 0),
        3:  (3, 4, 2, 0, 0, 0, 0, 0, 0, 0),
        4:  (4, 4, 3, 0, 0, 0, 0, 0, 0, 0),
        5:  (4, 4, 3, 2, 0, 0, 0, 0, 0, 0),
        6:  (4, 4, 3, 3, 0, 0, 0, 0, 0, 0),
        7:  (4, 4, 3, 3, 1, 0, 0, 0, 0, 0),
        8:  (4, 4, 3, 3, 2, 0, 0, 0, 0, 0),
        9:  (4, 4, 3, 3, 3, 1, 0, 0, 0, 0),
        10: (5, 4, 3, 3, 3, 2, 0, 0, 0, 0),
        11: (5, 4, 3, 3, 3, 2, 1, 0, 0, 0),
        12: (5, 4, 3, 3, 3, 2, 1, 0, 0, 0),
        13: (5, 4, 3, 3, 3, 2, 1, 1, 0, 0),
        14: (5, 4, 3, 3, 3, 2, 1, 1, 0, 0),
        15: (5, 4, 3, 3, 3, 2, 1, 1, 1, 0),
        16: (5, 4, 3, 3, 3, 2, 1, 1, 1, 0),
        17: (5, 4, 3, 3, 3, 2, 1, 1, 1, 1),
        18: (5, 4, 3, 3, 3, 3, 1, 1, 1, 1),
        19: (5, 4, 3, 3, 3, 3, 2, 1, 1, 1),
        20: (5, 4, 3, 3, 3, 3, 2, 2, 1, 1),
    }

    



def create_spell(**params):
    """Create a new subclass of ``Spell`` with given default parameters.
    
    Useful for spells that haven't been entered into the ``spells.py``
    file yet.
    
    Parameters
    ----------
    params : optional
      Saved as attributes of the new class.
    
    Returns
    -------
    NewSpell
      New spell class, subclass of ``Spell``, with given params.
    """
    NewSpell = Spell
    NewSpell.name = params.get('name', 'Unknown Spell')
    NewSpell.level = params.get('level', 9)
    return NewSpell


class Spell():
    """A magical spell castable by a player character."""
    level = 0
    name = "Unknown spell"
    casting_time = "1 action"
    casting_range = "60 ft"
    components = ()
    materials = ""
    duration = "instantaneous"
    ritual = False
    _concentration = False
    magic_school = ""
    classes = ()
    
    def __str__(self):
        if len(self.components) == 0:
            s = self.name
        else:
            s = self.name + ' ({:s}) '.format(','.join(self.components))
        # Indicate if this is a ritual or a concentration
        indicators = [('R', self.ritual), ('C', self.concentration), ('$', self.special_material)]
        indicators = tuple(letter for letter, is_active in indicators if is_active)
        if len(indicators):
            s += f' ({", ".join(indicators)})'
        return s
    
    def __repr__(self):
        return "\"{:s}\"".format(self.name)

    def __eq__(self, other):
        return (self.name == other.name) and (self.level == other.level)

    def __hash__(self):
        return 0
    
    @property
    def component_string(self):
        s = f'{", ".join(self.components)}'
        if "M" in self.components:
            s += f' ({self.materials})'
        return s
    
    @property
    def concentration(self):
        return ('concentration' in self.duration.lower()) or self._concentration

    @concentration.setter
    def concentration(self, val: bool):
        self._concentration = val

    @property
    def special_material(self):
        return ('worth at least' in self.materials.lower())




from .features import Feature
from .. import (weapons, armor, spells)


class UnarmoredDefenseMonk(Feature):
    """Beginning at 1st level, while you are wearing no armor and not wearing a
    shield, your AC equals 10 + your Dexterity modifier + your Wisdom modifier.

    This bonus is computed in the AC given on the Character Sheet above.

    """
    name = "Unarmored Defense"
    source = 'Monk'


class MartialArts(Feature):
    """At 1st level, your practice of martial arts gives you mastery of combat
    styles that use unarmed strikes and monk weapons, which are shortswords and
    any simple melee weapons that don’t have the two-handed or heavy
    property. You gain the following benefits while you are unarmed or wielding
    only monk weapons and you aren’t wearing armor or wielding a shield:

    -- You can use Dexterity instead of Strength for the attack and damage rolls
    of your unarmed strikes and monk weapons.

    -- You can roll a d4 in place of the normal damage of your unarmed strike or
    monk weapon. This die changes as you gain monk levels, as shown in the
    Martial Arts column of the Monk table.

    -- When you use the Attack action with an unarmed strike or a monk weapon on
    your turn, you can make one unarmed strike as a bonus action. For example,
    if you take the Attack action and attack with a quarter- staff, you can
    also make an unarmed strike as a bonus action, assuming you haven't already
    taken a bonus action this turn.

    Certain monasteries use specializepd forms of the monk weapons. For
    example, you might use a club that is two lengths of w ood connected by a
    short chain (called a nunchaku) or a sickle with a shorter, straighter
    blade (called a kama). Whatever name you use for a monk weapon, you can use
    the game statistics provided for

    """
    name = "Martial Arts"
    source = 'Monk'
    die = 'd4'

    def weapon_func(self, weapon: weapons.Weapon, **kwargs):
        """
        Update increasing damage dice and DEX mod of Monk weapons
        """
        is_monk_weapon = any([isinstance(weapon, w)
                              for w in weapons.monk_weapons])
        level = self.owner.Monk.level
        if not is_monk_weapon:
            return weapon
        self.die = 'd4'
        if level >= 5:
            self.die = 'd6'
        if level >= 11:
            self.die = 'd8'
        if level >= 17:
            self.die = 'd10'
        # check if new damage is better than default
        if int(self.die[1:]) > int(weapon.base_damage.split('d')[-1]):
            weapon.base_damage = '1' + str(self.die)
        weapon.is_finesse = True


class Ki(Feature):
    """Starting at 2nd level, your training allows you to harness the mystic
    energy of ki. Your access to this energy is represented by a number of ki
    points. Your monk level determines the number of points you have, as shown
    in the Ki Points column of the Monk table. You can spend these points to
    fuel various ki features.

    You start knowing three such features: Flurry of Blows, Patient Defense,
    and Step of the Wind. You learn more ki features as you gain levels in this
    class. When you spend a ki point, it is unavailable until you finish a
    short or long rest, at the end of which you draw all of your expended ki
    back into yourself. You must spend at least 30 minutes of the rest
    meditating to regain your ki points. 

    Some of your ki features require your target to make a saving throw to
    resist the feature’s effects. The saving throw DC is calculated as follows:
    Ki save DC = 8 + your proficiency bonus + your Wisdom modifier

    """
    _name = "Ki"
    source = "Monk"

    @property
    def name(self):
        num = self.owner.Monk.level
        DC = 8 + self.owner.proficiency_bonus + self.owner.wisdom.modifier
        return self._name + " ({:d} pts, DC={:d})".format(num, DC)


class FlurryOfBlows(Feature):
    """Immediately after you take the Attack action on your turn, you can spend 1
    ki point to make two unarmed strikes as a bonus action

    """
    name = "Flurry of Blows"
    source = "Monk"

    
class PatientDefense(Feature):
    """You can spend 1 ki point to take the Dodge action as a bonus action on your
    turn

    """
    name = "Patient Defense"
    source = "Monk"


class StepOfTheWind(Feature):
    """You can spend 1 ki point to take the Disengage or Dash action as a bonus
    action on your turn, and your jump distance is doubled for the turn

    """
    name = "Step of the Wind"
    source = "Monk"


class UnarmoredMovement(Feature):
    """Starting at 2nd level, your speed increases by 10 feet while you are not
    wearing armor or wielding a shield. This bonus increases when you reach
    certain monk levels, as shown in the Monk table.

    At 9th level, you gain the ability to move along vertical surfaces and
    across liquids on your turn without falling during the move.

    """
    name = "Unarmored Movement"
    source = "Monk"

    @property
    def speed_bonus(self):
        level = self.owner.Monk.level
        _speed_bonus = 10
        if level >= 6:
            _speed_bonus = 15
        if level >= 10:
            _speed_bonus = 20
        if level >= 14:
            _speed_bonus = 25
        if level >= 18:
            _speed_bonus = 30
        return _speed_bonus


class DeflectMissiles(Feature):
    """Starting at 3rd level, you can use your reaction to deflect or catch the
    missile when you are hit by a ranged weapon attack. When you do so, the
    damage you take from the attack is reduced by 1d10 + your Dexterity
    modifier + your monk level. If you reduce the damage to 0, you can catch
    the missile if it is small enough for you to hold in one hand and you have
    at least one hand free.

    If you catch a missile in this way, you can spend 1 ki point to make a
    ranged attack with the weapon or piece of ammunition you just caught, as
    part of the same reaction. You make this attack with proficiency,
    regardless of your weapon proficiencies, and the missile counts as a monk
    weapon for the attack

    """
    _name = "Deflect Missiles"
    source = "Monk"

    @property
    def name(self):
        mod = self.owner.dexterity.modifier + self.owner.Monk.level
        return self._name + " (1d10+{:d})".format(mod)


class SlowFall(Feature):
    """Beginning at 4th level, you can use your reaction when you fall to reduce
    any falling damage you take by an amount equal to five times your monk
    level.

    """
    name = "Slow Fall"
    source = "Monk"


class ExtraAttackMonk(Feature):
    """Beginning at 5th level, you can attack twice, instead of once, whenever you
    take the Attack action on your turn

    """
    name = "Extra Attack (2x)"
    source = "Monk"


class StunningStrike(Feature):
    """Starting at 5th level, you can interfere with the flow of ki in an
    opponent’s body. When you hit another creature with a melee weapon attack,
    you can spend 1 ki point to attempt a stunning strike. The target must
    succeed on a Constitution saving throw or be stunned until the end of your
    next turn

    """
    name = "Stunning Strike"
    source = "Monk"


class KiEmpoweredStrikes(Feature):
    """Starting at 6th level, your unarmed strikes count as magical for the
    purpose of overcoming resistance and immunity to nonmagical attacks and
    damage

    """
    name = "Ki-Empowered Strikes"
    source = "Monk"

    
class StillnessOfMind(Feature):
    """Starting at 7th level, you can use your action to end one effect on
    yourself that is causing you to be charmed or frightened

    """
    name = "Stillness of Mind"
    source = "Monk"


class PurityOfBody(Feature):
    """At 10th level, your mastery of the ki flowing through you makes you immune
    to disease and poison.

    """
    name = "Purity of Body"
    source = "Monk"


class TongueOfTheSunAndMoon(Feature):
    """Starting at 13th level, you learn to touch the ki of other minds so that
    you understand all spoken languages. Moreover, any creature that can
    understand a language can understand what you say.

    """
    name = "Tongue of the Sun and Moon"
    source = "Monk"


class DiamondSoul(Feature):
    """Beginning at 14th level, your mastery of ki grants you proficiency in all
    saving throws. Additionally, whenever you make a saving throw and fail, you
    can spend 1 ki point to reroll it and take the second result.

    """
    name = "Diamond Soul"
    source = "Monk"


class TimelessBody(Feature):
    """At 15th level, your ki sustains you so that you suffer none of the frailty
    of old age, and you can't be aged magically. You can still die of old age,
    however. In addition, you no longer need food or water.

    """
    name = "Timeless Body"
    source = "Monk"


class EmptyBody(Feature):
    """Beginning at 18th level, you can use your action to spend 4 ki points to
    become invisible for 1 minute. During that time, you also have resistance
    to all damage but force damage. Additionally, you can spend 8 ki points to
    cast the astral projection spell, without needing material components. When
    you do so, you can’t take any other creatures with you.

    """
    name = "Empty Body"
    source = "Monk"
        

class PerfectSelf(Feature):
    """At 20th level, when you roll for initiative and have no ki points
    remaining, you regain 4 ki points.

    """
    name = "Perfect Self"
    source = "Monk"


# Way of the Open Hand
class OpenHandTechnique(Feature):
    """Starting when you choose this tradition at 3rd level, you can manipulate
    your enemy’s ki when you harness your own. Whenever you hit a creature with
    one of the attacks granted by your Flurry of Blows, you can impose one of
    the following effects on that target:

    -- It must succeed on a Dexterity saving throw or be knocked prone.

    -- It must make a Strength saving throw. If it fails, you can push it up to
    15 feet away from you.

    -- It can’t take reactions until the end of your next turn

    """
    name = "Open Hand Technique"
    source = "Monk (Way of the Open Hand)"


class WholenessOfBody(Feature):
    """At 6th level, you gain the ability to heal yourself. As an action, you can
    regain hit points equal to three times your monk level. You must finish a
    long rest before you can use this feature again

    """
    name = "Wholeness of Body"
    source = "Monk (Way of the Open Hand)"


class Tranquility(Feature):
    """Beginning at 11th level, you can enter a special meditation that surrounds
    you with an aura of peace. At the end of a long rest, you gain the effect
    of a sanctuary spell that lasts until the start of your next long rest (the
    spell can end early as normal). The saving throw DC for the spell equals 8
    + your Wisdom modifier + your proficiency bonus

    """
    name = "Tranquility"
    source = "Monk (Way of the Open Hand)"


class QuiveringPalm(Feature):
    """At 17th level, you gain the ability to set up lethal vibrations in
    someone’s body. When you hit a creature with an unarmed strike, you can
    spend 3 ki points to start these imperceptible vibrations, which last for a
    number of days equal to your monk level. The vibrations are harmless unless
    you use your action to end them. To do so, you and the target must be on
    the same plane of existence. When you use this action, the creature must
    make a Constitution saving throw. If it fails, it is reduced to 0 hit
    points. If it succeeds, it takes 10d10 necrotic damage. You can have only
    one creature under the effect of this feature at a time. You can choose to
    end the vibrations harmlessly without using an action.

    """
    name = "Quivering Palm"
    source = "Monk (Way of the Open Hand)"


# Way of Shadow
class ShadowArts(Feature):
    """Starting when you choose this tradition at 3rd level, you can use your ki
    to duplicate the effects of certain spells. As an action, you can spend 2
    ki points to cast darkness, darkvision, pass without trace, or silence,
    without providing material components. Additionally, you gain the minor
    illusion cantrip if you don’t already know it.

    """
    name = "Shadow Arts"
    source = "Monk (Way of Shadow)"


class ShadowStep(Feature):
    """At 6th level, you gain the ability to step from one shadow into
    another. When you are in dim light or darkness, as a bonus action you can
    teleport up to 60 feet to an unoccupied space you can see that is also in
    dim light or darkness. You then have advantage on the first melee attack
    you make before the end of the turn.

    """
    name = "Shadow Step"
    source = "Monk (Way of Shadow)"


class CloakOfShadows(Feature):
    """By 11th level, you have learned to become one with the shadows. When you
    are in an area of dim light or darkness, you can use your action to become
    invisible. You remain invisible until you make an attack, cast a spell, or
    are in an area of bright light.

    """
    name = "Cloak of Shadows"
    source = "Monk (Way of Shadow)"


class Opportunist(Feature):
    """At 17th level, you can exploit a creature's momentary distraction when it
    is hit by an attack. Whenever a creature within 5 feet of you is hit by an
    attack made by a creature other than you, you can use your reaction to make
    a melee attack against that creature.

    """
    name = "Opportunist"
    source = "Monk (Way of Shadow)"


# Way of the Four Elements
class DiscipleOfTheElements(Feature):
    """When you choose this tradition at 3rd level, you learn magical disciplines
    that harness the power of the four elements. A discipline requires you to
    spend ki points each time you use it. You know the Elemental Attunement
    discipline and one other elemental discipline of your choice, which are
    detailed in the “Elemental Disciplines” section below.

    You learn one additional elemental discipline of your choice at 6th, 11th,
    and 17th level. Whenever you learn a new elemental discipline, you can also
    replace one elemental discipline that you already know with a different
    discipline. 

    Add your chosen disciplines under "features" in your .py file
    
    **Casting Elemental Spells**: Some elemental disciplines allow you to cast
    spells. See chapter 10 for the general rules of spellcasting. To cast one o
    f these spells, you use its casting time and other rules, but you don’t
    need to provide material components for it. Once you reach 5th level in
    this class, you can spend additional ki points to increase the level of an
    elemental discipline spell that you cast, provided that the spell has an
    enhanced effect at a higher level, as burning hands does. The spell's level
    increases by 1 for each additional ki point you spend. For example, if you
    are a 5th-level monk and use Sweeping Cinder Strike to cast burning hands,
    you can spend 3 ki points to cast it as a 2nd-level spell (the discipline’s
    base cost of 2 ki points plus 1).

    The maximum number of ki points you can spend to cast a spell in this way
    (including its base ki point cost and any additional ki points you spend to
    increase its level) is determined by your monk level, as shown in the
    Spells and Ki Points table.

    Monk Levels 5-8 : 3 Ki points Max

    Monk Levels 9-12 : 4 Ki points Max

    Monk Levels 13-16 : 5 Ki points Max

    Monk Levels 17-20 : 6 Ki points Max

    """
    name = "Disciple of the Elements"
    source = "Monk (Way of the Four Elements)"


class ElementalAttunement(Feature):
    """You can use your action to briefly control elemental forces nearby, causing
    one of the following effects of your choice:

    -- Create a harmless, instantaneous sensory effect related to air, earth,
    fire, or water, such as a shower of sparks, a puff of wind, a spray o f
    light mist, or a gentle rumbling of stone.

    -- Instantaneously light or snuff out a candle, a torch, or a small
    campfire.

    -- Chill or warm up to 1 pound of nonliving material for up to 1 hour.

    -- Cause earth, fire, water, or mist that can fit within a 1-foot cube to
    shape itself into a crude form you desig nate for 1 minute.

    """
    name = "Elemental Attunement"
    source = "Monk (Way of the Four Elements)"


class BreathOfWinter(Feature):
    """You can spend 6 ki points to cast cone of cold.

    **Prerequisite**: 17th Level

    """
    name = "Breath of Winter"
    source = "Monk (Way of the Four Elements)"
    spells_known = (spells.ConeOfCold,)


class ClenchOfTheNorthWind(Feature):
    """You can spend 3 ki points to cast hold person.
    
    **Prerequisite**: 6th Level

    """
    name = "Clench of the North Wind"
    source = "Monk (Way of the Four Elements)"
    spells_known = (spells.HoldPerson,)


class EternalMountainDefense(Feature):
    """You can spend 5 ki points to cast stoneskin, targeting yourself.

    **Prerequisite**: 11th Level
    """
    name = "Eternal Mountain Defense"
    source = "Monk (Way of the Four Elements)"
    spells_known = (spells.Stoneskin,)

    
class FangsOfTheFireSnake(Feature):
    """When you use the Attack action on your turn, you can spend 1 ki point to
    cause tendrils of flame to stretch out from your fists and feet. Your reach
    with your unarmed strikes increases by 10 feet for that action, as well as
    the rest o f the turn. A hit with such an attack deals fire damage instead
    of bludgeoning damage, and if you spend 1 ki point when the attack hits, it
    also deals an extra 1d10 fire damage

    """
    name = "Fangs of the Fire Snake"
    source = "Monk (Way of the Four Elements)"


class FistOfFourThunders(Feature):
    """You can spend 2 ki points to cast thunderwave

    """
    name = "Fist of Four Thunders"
    source = "Monk (Way of the Four Elements)"
    spells_known = (spells.Thunderwave,)


class FistOfUnbrokenAir(Feature):
    """You can create a blast of compressed air that strikes like a mighty
    fist. As an action, you can spend 2 ki points and choose a creature within
    30 feet of you. That creature must make a Strength saving throw. On a
    failed save, the creature takes 3d10 bludgeoning damage, plus an extra 1d10
    bludgeoning damage for each additional ki point you spend, and you can push
    the creature up to 20 feet away from you and knock it prone. On a
    successful save, the creature takes half as much damage, and you don’t push
    it or knock it prone.

    """
    name = "Fist of Unbroken Air"
    source = "Monk (Way of the Four Elements)"


class FlamesOfThePhoenix(Feature):
    """You can spend 4 ki points to cast fireball.

    **Prerequisite**: 11th Level

    """
    name = "Flames of the Phoenix"
    source = "Monk (Way of the Four Elements)"
    spells_known = (spells.Fireball,)


class GongOfTheSummit(Feature):
    """You can spend 3 ki points to cast shatter.

    **Prerequisite**: 6th Level

    """
    name = "Gong of the Summit"
    source = "Monk (Way of the Four Elements)"
    spells_known = (spells.Shatter,)


class MistStance(Feature):
    """You can spend 4 ki points to cast gaseous form, targeting yourself.

    **Prerequisite**: 11th Level

    """
    name = "Mist Stance"
    source = "Monk (Way of the Four Elements)"
    spells_known = (spells.GaseousForm,)


class RideTheWind(Feature):
    """You can spend 4 ki points to cast fly, targeting yourself

    **Prerequisite**: 11th Level

    """
    name = "Ride the Wind"
    source = "Monk (Way of the Four Elements)"
    spells_known = (spells.Fly,)


class RiverOfHungryFlame(Feature):
    """You can spend 5 ki points to cast wall of fire.

    **Prerequisite**: 17th Level

    """
    name = "River of Hungry Flame"
    source = "Monk (Way of the Four Elements)"
    spells_known = (spells.WallOfFire,)


class RushOfTheGaleSpirits(Feature):
    """You can spend 2 ki points to cast gust of wind.

    """
    name = "Rush of the Gale Spirits"
    source = "Monk (Way of the Four Elements)"
    spells_known = (spells.GustOfWind,)


class ShapeTheFlowingRiver(Feature):
    """As an action, you can spend 1 ki point to choose an area of ice or water no
    larger than 30 feet on a side within 120 feet o f you. You can change water
    to ice within the area and vice versa, and you can reshape ice in the area
    in any manner you choose. You can raise or lower the ice’s elevation,
    create or fill in a trench, erect or flatten a wall, or form a pillar. The
    extent of any such changes can’t exceed half the area’s largest
    dimension. For example, if you affect a 30-foot square, you can create a
    pillar up to 15 feet high, raise or lower the square’s elevation by up to
    15 feet, dig a trench up to 15 feet deep, and so on. You can’t shape the
    ice to trap or injure a creature in the area.

    """
    name = "Shape the Flowing River"
    source = "Monk (Way of the Four Elements)"


class SweepingCinderStrike(Feature):
    """You can spend 2 ki points to cast burning hands

    """
    name = "Sweeping Cinder Strike"
    source = "Monk (Way of the Four Elements)"
    spells_known = (spells.BurningHands,)


class WaterWhip(Feature):
    """You can spend 2 ki points as a bonus action to create a whip of water that
    shoves and pulls a creature to unbalance it. A creature that you can see
    that is within 30 feet of you must make a Dexterity saving throw. On a
    failed save, the creature takes 3d10 bludgeoning damage, plus an extra 1d10
    bludgeoning damage for each additional ki point you spend, and you can
    either knock it prone or pull it up to 25 feet closer to you. On a
    successful save, the creature takes half as much damage, and you don’t pull
    it or knock it prone

    """
    name = "Water Whip"
    source = "Monk (Way of the Four Elements)"


class WaveOfRollingEarth(Feature):
    """You can spend 6 ki points to cast wall of stone

    **Prerequisite**: 17th Level

    """
    name = "Wave of Rolling Earth"
    source = "Monk (Way of the Four Elements)"
    spells_known = (spells.WallOfStone,)


# Way of the Long Death
class TouchOfDeath(Feature):
    """Starting when you choose this tradition at 3rd level, your study of death
    allows you to extract vitality from an- other creature as it nears its
    demise. When you reduce a creature within 5 feet of you to 0 hit points,
    you gain temporary hit points equal to your Wisdom modifier + your monk
    level (minimum of 1 temporary hit point)

    """
    name = 'Touch of Death'
    source = "Monk (Way of the Sun Soul)"


class HourOfReaping(Feature):
    """At 6th level, you gain the ability to unsettle or terrify those around you
    as an action, for your soul has been touched by the shadow of death. When
    you take this ac- tion , each creature within 30 feet of you that can see
    you must succeed on a Wisdom saving throw or be fright- ened of you until
    the end of your next turn

    """
    name = "Hour of Reaping"
    source = "Monk (Way of the Sun Soul)"


class MasteryOfDeath(Feature):
    """Beginning at 11th level, you use your familiarity with death to escape its
    grasp. When you are reduced to 0 hit points, you can expend 1 ki point (no
    action required) to have 1 hit point instead

    """
    name = "Mastery of Death"
    source = "Monk (Way of the Sun Soul)"
        

class TouchOfTheLongDeath(Feature):
    """Starting at 17th level, your touch can channel the energy of death into a
    creature. As an action, you touch one creature within 5 feet of you, and
    you expend 1 to 10 ki points. The target must make a Constitution saving
    throw, and it takes 2d10 necrotic damage per ki point spent on a failed
    save, or half as much damage on a suc- cessful one

    """
    name = "Touch of the Long Death"
    source = "Monk (Way of the Sun Soul)"
        

# Way of the Sun Soul
class RadiantSunBolt(Feature):
    """Starting when you choose this tradition at 3rd level, you can hurl searing
    bolts of magical radiance. You gain a ranged spell attack that you can use
    with the Attack action. The attack has a range of 30 feet. You are
    proficient with it, and you add your Dexterity modi- fier to its attack and
    damage rolls. Its damage is radiant, and its damage die is a d4.

    This die changes as you gain monk levels, as shown in the Martial Arts
    column of the Monk table. When you use the Attack action on your turn to
    use this special attack, you can spend 1 ki point to make two additional
    attacks with it as a bonus action.

    """
    name = "Radiant Sun Bolt"
    source = "Monk (Way of the Sun Soul)"

    def __init__(self, owner=None):
        super().__init__(owner=owner)
        self.owner.wield_weapon("sun bolt")
    

class SearingArcStrike(Feature):
    """At 6th level, you gain the ability to channel your ki into searing waves of
    energy. Immediately after you take the Attack action on your turn, you can
    spend 2 ki points to cast the 1st-level spell burning hands as a bonus
    action. You can spend additional ki points to cast burning hands as a
    higher level spell. Each additional ki point you spend increases the
    spell's level by 1. The maximum number of ki points (2 plus any additional
    points) that you can spend on the spell equals half your monk level (round
    down)

    """
    name = "Searing Arc Strike"
    source = "Monk (Way of the Sun Soul)"
    spells_known = (spells.BurningHands,)


class SearingSunburst(Feature):
    """At 11th level, you gain the ability to create an orb of light that erupts
    into a devastating explosion. As an action, you create an orb and hurl it
    at a point you choose within 150 feet, where it erupts into a sphere of
    radiant light for a brief but deadly instant. Each creature in that
    20-foot-radius sphere must succeed on a Constitution saving throw or take
    2d6 radiant damage. A creature doesn't need to make the save if the
    creature is behind total cover that is opaque. You can increase the
    sphere's damage by spending ki points. Each point you spend, up to a
    maximum of 3, increases the damage by 2d6.

    """
    name = "Searing Sunburst"
    source = "Monk (Way of the Sun Soul)"


class SunShield(Feature):
    """At 17th level, you become wreathed in a luminous aura. You shed bright
    light in a 30-foot radius and dim light for an additional 30 feet. You can
    extinguish or restore the light as a bonus action. If a creature hits you
    with a melee attack while this light shines, you can use you r reaction to
    deal radiant damage to the creature. The radiant damage equals 5 + your
    Wisdom modifier

    """
    name = "Sun Shield"
    source = "Monk (Way of the Sun Soul)"
        

# Way of the Drunken Master
class DrunkenTechnique(Feature):
    """At 3rd level, you learn how to twist and turn quickly as part of your
    Flurry of Blows. Whenever you use Flurry of Blows, you gain the benefit of
    the Disengage action, and your walking speed increases by 10 feet until the
    end of the current turn

    """
    name = "Drunken Technique"
    source = "Monk (Way of the Drunken Master)"


class TipsySway(Feature):
    """Starting at 6th level, you can move in sudden, swaying ways. You gain the
    following benefits.

    **Leap to Your Feet**: When you’re prone, you can stand up
    by spending 5 feet of movement, rather than half your speed.

    **Redirect Attack**: When a creature misses you with a melee attack roll,
    you can spend 1 ki point as a re- action to cause that attack to hit one
    creature of your choice, other than the attacker, that you can see within 5
    feet of you.

    """
    name = "Tipsy Sway"
    source = "Monk (Way of the Drunken Master)"


class DrunkardsLuck(Feature):
    """Starting at llth level, you always seem to get a lucky bounce at the right
    moment. When you make an ability check, an attack roll, or a saving throw
    and have disad- vantage on the roll, you can spend 2 ki points to cancel
    the disadvantage for that roll

    """
    name = "Drunkard's Luck"
    source = "Monk (Way of the Drunken Master)"


class IntoxicatedFrenzy(Feature):
    """At 17th level, you gain the ability to make an overwhelm- ing number of
    attacks against a group of enemies. When you use your Flurry of Blows, you
    can make up to three additional attacks with it (up to a total of five
    Flurry of Blows attacks), provided that each Flurry of Blows at- tack
    targets a different creature this turn

    """
    name = "Intoxicated Frenzy"
    source = "Monk (Way of the Drunken Master)"


# Way of the Kensei
class PathOfTheKensei(Feature):
    """When you choose this tradition at 3rd level, your spe- cial martial arts
    training leads you to master the use of certain weapons. This path also
    includes instruction in the deft strokes of calligraphy or painting. You
    gain the following benefits.

    **Kensei Weapons**: Choose two types of weapons to be your kensei weapons:
    one melee weapon and one ranged weapon. Each of these weapons can be any
    sim- ple or martial weapon that lacks the heavy and special properties. The
    longbow is also a valid choice. You gain proficiency with these weapons if
    you don’t already have it. Weapons of the chosen types are monk weapons for
    you. Many of this tradition’s features work only with your kensei
    weapons. When you reach 6th, 11th, and 17th level in this class. you can
    choose another type of weapon—either melee or ranged—to be a kensei weapon
    for you. following the criteria above.

    **Agile Parry**: If you make an unarmed strike as part of the Attack action
    on your turn and are holding a kensei weapon, you can use it to defend
    yourself if it is a melee weapon. You gain a +2 bonus to AC until the start
    of your next turn, while the weapon is in your hand and you aren't
    incapacitated.

    **Kensei’s Shot**: You can use a bonus action on your turn to make your
    ranged attacks with a kensei weapon more deadly. When you do so, any target
    you hit with a ranged attack using a kensei weapon takes an extra 1d4
    damage of the weapons type. You retain this benefit un— til the end of the
    current turn.

    **Way ofthe Brush**: You gain proficiency with your choice of
    calligrapher's supplies or painter's supplies

    """
    name = "Path of the Kensei"
    source = "Monk (Way of the Kensei)"


class OneWithTheBlade(Feature):
    """At 6th level, you extend your ki into your kensei weap- ons, granting you
    the following benefits.

    **Magic Kensei Weapons**: Your attacks with your kensei weapons count as
    magical for the purpose of over- coming resistance and immunity to
    nonmagical attacks and damage

    **Deft Strike**: When you hit a target with a kensei weapon, you can spend
    1 ki point to cause the weapon to deal extra damage to the target equal to
    your Martial Arts die. You can use this feature only once on each of
    your turns.

    """
    name = "One with the Blade"
    source = "Monk (Way of the Kensei)"


class SharpenTheBlade(Feature):
    """At 11th level, you gain the ability to augment your weap- ons further with
    your ki. As a bonus action, you can expend up to 3 ki points to grant one
    kensei weapon you touch a bonus to attack and damage rolls when you attack
    with it. The bonus equals the number of ki points you spent. This bonus
    lasts for 1 minute or until you use this feature again. This feature has no
    effect on a magic weapon that already has a bonus to attack and damage
    rolls

    """
    name = "Sharpen the Blade"
    source = "Monk (Way of the Kensei)"


class UnerringAccuracy(Feature):
    """At 17th level, your mastery of weapons grants you ex- traordinary
    accuracy. Ifyou miss with an attack roll using a monk weapon on your turn,
    you can reroll it. You can use this feature only once on each of your
    turns.

    """
    name = "Unerring Accuracy"
    source = "Monk (Way of the Kensei)"




__all__ = ('Monk')

from .. import (features, weapons)
from .classes import CharClass, SubClass
from collections import defaultdict


# PHB
class OpenHandWay(SubClass):
    """Monks of the Way of the Open Hand are the ultimate masters of martial arts
    combat, whether armed or unarmed. They learn techniques to push and trip
    their opponents, manipulate ki to heal damage to their bodies, and practice
    advanced meditation that can protect them from harm.

    """
    name = "Way of the Open Hand"
    features_by_level = defaultdict(list)
    features_by_level[3] = [features.OpenHandTechnique]
    features_by_level[6] = [features.WholenessOfBody]
    features_by_level[11] = [features.Tranquility]
    features_by_level[17] = [features.QuiveringPalm]


class ShadowWay(SubClass):
    """Monks of the Way of Shadow follow a tradition that values stealth and
    subterfuge. These monks might be called ninjas or shadowdancers, and they
    serve as spies and assassins. Sometimes the members of a ninja monastery
    are family members, forming a clan sworn to secrecy about their arts and
    missions. Other monasteries are more like thieves’ guilds, hiring out their
    services to nobles, rich merchants, or anyone else who can pay their
    fees. Regardless of their methods, the heads of these monasteries expect
    the unquestioning obedience of their students

    """
    name = "Way of Shadow"
    features_by_level = defaultdict(list)
    features_by_level[3] = [features.ShadowArts]
    features_by_level[6] = [features.ShadowStep]
    features_by_level[11] = [features.CloakOfShadows]
    features_by_level[17] = [features.Opportunist]
    

class FourElementsWay(SubClass):
    """You follow a monastic tradition that teaches you to harness the
    elements. When you focus your ki, you can align yourself with the forces of
    creation and bend the four elements to your will, using them as an
    extension of your body. Some members of this tradition dedicate themselves
    to a single element, but others weave the elements together.

    Many monks of this tradition tattoo their bodies with representations of
    their ki powers, commonly imagined as coiling dragons, but also as
    phoenixes, fish, plants, mountains, and cresting waves.

    """
    name = "Way of the Four Elements"
    spellcasting_ability = 'wisdom'
    features_by_level = defaultdict(list)
    features_by_level[3] = [features.DiscipleOfTheElements,
                            features.ElementalAttunement]


# SCAG
class LongDeathWay(SubClass):
    """Monks of the Way of the Long Death are obsessed with the meaning and
    mechanics of dying. They capture creatures and prepare elaborate
    experiments to capture, record, and understand the moments of their
    demise. They then use this knowledge to guide their understanding of
    martial arts, yielding a deadly fighting style.

    """
    name = "Way of the Long Death"
    features_by_level = defaultdict(list)
    features_by_level[3] = [features.TouchOfDeath]
    features_by_level[6] = [features.HourOfReaping]
    features_by_level[11] = [features.MasteryOfDeath]
    features_by_level[17] = [features.TouchOfTheLongDeath]


class SunSoulWay(SubClass):
    """Monks of the Way of the Sun Soul learn to channel their own life energy
    into searing bolts of light. They teach that meditation can unlock the
    ability to unleash the indomitable light shed by the soul of every living
    creature

    """
    name = "Way of the Sun Soul"
    features_by_level = defaultdict(list)
    features_by_level[3] = [features.RadiantSunBolt]
    features_by_level[6] = [features.SearingArcStrike]
    features_by_level[11] = [features.SearingSunburst]
    features_by_level[17] = [features.SunShield]


# XGTE
class DrunkenMasterWay(SubClass):
    """The Way of the Drunken Master teaches its students to move with the jerky,
    unpredictable movements of a drunkard. A drunken master sways, tottering on
    unsteady feet, to present what seems like an incompetent combatant who
    proves frustrating to engage. The drunken master’s erratic stumbles conceal
    a carefully executed dance of blocks, parries, advances, attacks, and
    retreats.

    A drunken master often enjoys playing the fool to bring gladness to the
    despondent or to demonstrate humility to the arrogant, but when battle is
    joined, the drunken master can be a maddening, masterful foe

    """
    name = "Way of the Drunken Master"
    _proficiencies_text = ("brewer's suplies",)
    skill_proficiencies = ("performance",)
    features_by_level = defaultdict(list)
    features_by_level[3] = [features.DrunkenTechnique]
    features_by_level[6] = [features.TipsySway]
    features_by_level[11] = [features.DrunkardsLuck]
    features_by_level[17] = [features.IntoxicatedFrenzy]


class KenseiWay(SubClass):
    """Monks of the Way of the Kensei train relentlessly with their weapons, to
    the point where the weapon becomes an extension of the body. Founded on a
    mastery of sword fighting, the tradition has expanded to include many
    different weapons.

    A kensei sees a weapon in much the same way a calligrapher or painter
    regards a pen or brush. Whatever the weapon, the kensei views it as a tool
    used to express the beauty and precision of the martial arts. That such
    mastery makes a kensei a peerless warrior is but a side effect of intense
    devotion, practice, and study.

    """
    name = "Way of the Kensei"
    features_by_level = defaultdict(list)
    features_by_level[3] = [features.PathOfTheKensei]
    features_by_level[6] = [features.OneWithTheBlade]
    features_by_level[11] = [features.SharpenTheBlade]
    features_by_level[17] = [features.UnerringAccuracy]


class Monk(CharClass):
    name = 'Monk'
    hit_dice_faces = 8
    subclass_select_level = 3
    saving_throw_proficiencies = ('strength', 'dexterity')
    primary_abilities = ('dexterity', 'wisdom')
    _proficiencies_text = (
        'simple weapons', 'shortswords', 'unarmed',
        "one type of artisan's tools or one musical instrument")
    weapon_proficiencies = (weapons.Shortsword, weapons.Unarmed,
                            weapons.SimpleWeapon)
    multiclass_weapon_proficiencies = weapon_proficiencies
    _multiclass_proficiencies_text = ('simple weapons', 'shortswords',
                                      'unarmed')
    class_skill_choices = ('Acrobatics', 'Athletics', 'History', 'Insight',
                           'Religion', 'Stealth')
    subclasses_available = (OpenHandWay, ShadowWay,
                            FourElementsWay, SunSoulWay,
                            LongDeathWay, DrunkenMasterWay,
                            KenseiWay)
    features_by_level = defaultdict(list)
    features_by_level[1] = [features.UnarmoredDefenseMonk,
                            features.MartialArts]
    features_by_level[2] = [features.Ki, features.FlurryOfBlows,
                            features.PatientDefense, features.StepOfTheWind,
                            features.UnarmoredMovement]
    features_by_level[3] = [features.DeflectMissiles]
    features_by_level[4] = [features.SlowFall]
    features_by_level[5] = [features.ExtraAttackMonk, features.StunningStrike]
    features_by_level[6] = [features.KiEmpoweredStrikes]
    features_by_level[7] = [features.Evasion, features.StillnessOfMind]
    features_by_level[10] = [features.PurityOfBody]
    features_by_level[13] = [features.TongueOfTheSunAndMoon]
    features_by_level[14] = [features.DiamondSoul]
    features_by_level[15] = [features.TimelessBody]
    features_by_level[18] = [features.EmptyBody]
    features_by_level[20] = [features.PerfectSelf]

    def __init__(self, level, subclass=None, **params):
        super().__init__(level, subclass=subclass, **params)
        for f in self.features_by_level[1]:
            if isinstance(f, features.MartialArts):
                f.level = self.level



from .features import Feature
from .. import spells


# PHB
class FontOfMagic(Feature):
    """At 2nd level, you tap into a deep wellspring of magic within yourself. This
    wellspring is represented by sorcery points, which allow you to create a
    variety of magical effects.

    **Sorcery Points**: You have sorcery points equal to your Sorceror
    Level. You can never have more sorcery points than shown on the table for
    your level. You regain all spent sorcery points when you finish a long
    rest.

    **Flexible Casting**: You can use your sorcery points to gain additional
    spell slots, or sacrifice spell slots to gain additional sorcery
    points. You learn other ways to use your sorcery points as you reach higher
    levels. You can transform unexpended sorcery points into one spell slot as
    a bonus action on your turn. The Creating Spell Slots table shows the cost
    of creating a spell slot of a given level. You can create spell slots no
    higher in level than 5th. As a bonus action on your turn, you can expend
    one spell slot and gain a number of sorcery points equal to the slot’s
    level.

    1st Level Slot <--> 2 sorcery points

    2nd Level Slot <--> 3 sorcery points

    3rd Level Slot <--> 5 sorcery points

    4th Level Slot <--> 6 sorcery points

    5th Level Slot <--> 7 sorcery points

    """
    name = "Font of Magic"
    source = "Sorceror"


class Metamagic(Feature):
    """At 3rd level, you gain the ability to twist your spells to suit your
    needs. You gain two of the following Metamagic options of your choice. You
    gain another one at 10th and 17th level. You can use only one Metamagic
    option on a spell when you cast it, unless otherwise noted

    """
    name = "Metamagic"
    source = "Sorceror (Metamagic)"


class SorcerousRestoration(Feature):
    """At 20th level, you regain 4 expended sorcery points whenever you finish a
    short rest.

    """
    name = "Sorcerous Restoration"
    source = "Sorceror"


# Metamagic
class CarefulSpell(Metamagic):
    """When you cast a spell that forces other creatures to make a saving throw,
    you can protect some of those creatures from the spell’s full force. To do
    so, you spend 1 sorcery point and choose a number o f those creatures up to
    your Charisma modifier (minimum of one creature). A chosen creature
    automatically succeeds on its saving throw against the spell.

    """
    name = "Careful Spell"


class DistantSpell(Metamagic):
    """When you cast a spell that has a range of 5 feet or greater, you can spend
    1 sorcery point to double the range of the spell. When you cast a spell
    that has a range of touch, you can spend 1 sorcery point to make the range
    of the spell 30 feet

    """
    name = "Distant Spell"


class EmpoweredSpell(Metamagic):
    """When you roll damage for a spell, you can spend 1 sorcery point to reroll a
    number of the damage dice up to your Charisma modifier (minimum of
    one). You must use the new rolls. You can use Empowered Spell even if you
    have already used a different Metamagic option during the casting of the
    spell.

    """
    name = "Empowered Spell"


class ExtendedSpell(Metamagic):
    """When you cast a spell that has a duration of 1 minute or longer, you can
    spend 1 sorcery point to double its duration, to a maximum duration of 24
    hours.

    """
    name = "Extended Spell"


class HeightenedSpell(Metamagic):
    """When you cast a spell that forces a creature to make a saving throw to
    resist its effects, you can spend 3 sorcery points to give one target of
    the spell disadvantage on its first saving throw made against the spell

    """
    name = "Heightened Spell"


class QuickenedSpell(Metamagic):
    """When you cast a spell that has a casting time of 1 action, you can spend 2
    sorcery points to change the casting time to 1 bonus action for this
    casting.

    """
    name = "Quickened Spell"


class SubtleSpell(Metamagic):
    """When you cast a spell, you can spend 1 sorcery point to cast it without any
    somatic or verbal components.

    """
    name = "Subtle Spell"


class TwinnedSpell(Metamagic):
    """When you cast a spell that targets only one creature and doesn’t have a
    range of self, you can spend a number of sorcery points equal to the
    spell’s level to target a second creature in range with the same spell (1
    sorcery point if the spell is a cantrip)

    """
    name = "Twinned Spell"


# Wild Magic
class WildMagicSurge(Feature):
    """Starting when you choose this origin at 1st level, your spellcasting can
    unleash surges of untamed magic. Immediately after you cast a sorcerer
    spell of 1st level or higher, the DM can have you roll a d20. If you roll a
    1, roll on the Wild Magic Surge table to create a random magical effect.

    """
    name = "Wild Magic Surge"
    source = "Sorceror (Wild Magic)"


class TidesOfChaos(Feature):
    """Starting at 1st level, you can manipulate the forces of chance and chaos to
    gain advantage on one attack roll, ability check, or saving throw. Once you
    do so, you must finish a long rest before you can use this feature again.

    Any time before you regain the use of this feature, the DM can have you
    roll on the Wild Magic Surge table immediately after you cast a sorcerer
    spell of 1st level or higher. You then regain the use of this feature.

    """
    name = "Tides of Chaos"
    source = "Sorceror (Wild Magic)"


class BendLuck(Feature):
    """Starting at 6th level, you have the ability to twist fate using your wild
    magic. When another creature you can see makes an attack roll, an ability
    check, or a saving throw, you can use your reaction and spend 2 sorcery
    points to roll 1d4 and apply the number rolled as a bonus or penalty (your
    choice) to the creature’s roll. You can do so after the creature rolls but
    before any effects of the roll occur.

    """
    name = 'Bend Luck'
    source = 'Sorceror (Wild Magic)'


class ControlledChaos(Feature):
    """At 14th level, you gain a modicum of control over the surges of your wild
    magic. Whenever you roll on the Wild Magic Surge table, you can roll twice
    and use either number.

    """
    name = 'Controlled Chaos'
    source = "Sorceror (Wild Magic)"


class SpellBombardment(Feature):
    """Beginning at 18th level, the harmful energy of your spells
    intensifies. When you roll damage for a spell and roll the highest number
    possible on any of the dice, choose one of those dice, roll it again and
    add that roll to the damage. You can use the feature only once per turn.

    """
    name = "Spell Bombardment"
    source = "Sorceror (Wild Magic)"


# Draconic Ancestry
class DraconicResilience(Feature):
    """As magic flows through your body, it causes physical traits of your dragon
    ancestors to emerge. At 1st level, your hit point maximum increases by 1
    and increases by 1 again whenever you gain a level in this class.

    Additionally, parts of your skin are covered by a thin sheen of dragon-like
    scales. When you aren’t wearing armor, your AC equals 13 + your Dexterity
    modifier.

    This bonus is computed in the AC given on the Character Sheet above.

    """
    name = "Draconic Resilience"
    source = "Sorceror (Draconic Bloodline)"


class DragonAncestor(Feature):
    """At 1st level, you choose one type of dragon as your ancestor. The damage
    type associated with each dragon is used by features you gain later

    Dragon : Damage
    
    Black : Acid
    
    Blue : Lightning

    Brass : Fire

    Bronze : Lightning

    Copper : Acid

    Gold : Fire

    Green : Poison

    Red : Fire

    Silver : Cold
    
    White : Cold

    You can speak, read, and write Draconic. Additionally, whenever you make a
    Charisma check when interacting with dragons, your proficiency bonus is
    doubled if it applies to the check.

    """
    name = "Dragon Ancestor"
    source = "Sorceror (Draconic Bloodline)"


class ElementalAffinity(Feature):
    """Starting at 6th level, when you cast a spell that deals damage of the type
    associated with your draconic ancestry, add your Charisma modifier to that
    damage. At the same time, you can spend 1 sorcery point to gain resistance
    to that damage type for 1 hour

    """
    name = "Elemental Affinity"
    source = "Sorceror (Draconic Bloodline)"

    
class DragonWings(Feature):
    """At 14th level, you gain the ability to sprout a pair of dragon wings from
    your back, gaining a flying speed equal to your current speed. You can
    create these wings as a bonus action on your turn. They last until you
    dismiss them as a bonus action on your turn. You can’t manifest your wings
    while wearing armor unless the armor is made to accommodate them, and
    clothing not made to accommodate your wings might be destroyed when you
    manifest them

    """
    name = "Dragon Wings"
    source = "Sorceror (Draconic Bloodline)"


class DraconicPresence(Feature):
    """Beginning at 18th level, you can channel the dread presence of your dragon
    ancestor, causing those around you to become awestruck or frightened. As an
    action, you can spend 5 sorcery points to draw on this power and exude an
    aura of awe or fear (your choice) to a distance of 60 feet. For 1 minute or
    until you lose your concentration (as if you were casting a concentration
    spell), each hostile creature that starts its turn in this aura must
    succeed on a Wisdom saving throw or be charmed (if you chose awe) or
    frightened (if you chose fear) until the aura ends. A creature that
    succeeds on this saving throw is immune to your aura for 24 hours.

    """
    name = "Draconic Presence"
    source = "Sorceror (Draconic Bloodline)"


# Divine Soul
class DivineMagic(Feature):
    """Your link to the divine allows you to learn spells from the cleric
    class. When your Spellcasting feature lets you learn or replace a sorcerer
    cantrip or a sorcerer spell of 1st level or higher, you can choose the new
    spell from the cleric spell list or the sorcerer spell list. You must
    otherwise obey all the restrictions for selecting the spell, and it becomes
    a sorcerer spell for you.

    In addition, choose an affinity for the source of your divine power: good,
    evil, law, chaos, or neutrality. You learn an additional spell based on
    that affinity, as shown below. It is a sorcerer spell for you, but it
    doesn’t count against your number of sorcerer spells known. If you later
    replace this spell, you must replace it with a Spell from the cleric spell
    list

    Good : Cure Wounds

    Evil : Inflict Wounds

    Law : Bless

    Chaos : Bane

    Neutrality : Protection from Evil and Good

    """
    name = "Divine Magic"
    source = "Sorceror (Divine Soul)"


class FavoredByTheGods(Feature):
    """Starting at 1st level, divine power guards your destiny. If you fail a
    saving throw or miss with an attack roll, you can roll 2d4 and add it to
    the total, possibly changing

    """
    name = "Favored by the Gods"
    source = "Sorceror (Divine Soul)"


class EmpoweredHealing(Feature):
    """Starting at 6th level, the divine energy coursing through you can empower
    healing spells. Whenever you or an ally within 5 feet of you rolls dice to
    determine the number of hit points a spell restores, you can spend 1
    sorcery point to reroll any number of those dice once, provided you aren't
    incapacitated. You can use this feature only once per turn.

    """
    name = "Empowered Healing"
    source = "Sorceror (Divine Soul)"


class OtherworldlyWings(Feature):
    """Starting at 14th level, you can use a bonus action to manifest a pair of
    spectral wings from your back. While the wings are present, you have a
    flying speed of 30 feet. The wings last until you're incapacitated, you
    die, or you dismiss them as a bonus action. The affinity you chose for your
    Divine Magic feature determines the appearance of the spectral wings: eagle
    wings for good or law, bat wings for evil or chaos, and dragonfly wings for
    neutrality

    """
    name = "Otherworldly Wings"
    source = "Sorceror (Divine Soul)"


class UnearthlyRecovery(Feature):
    """At 18th level, you gain the ability to overcome grievous injuries. As a
    bonus action when you have fewer than half of your hit points remaining,
    you can regain a number of hit points equal to half your hit point
    maximum. Once you use this feature, you can’t use it again until you finish
    a long rest

    """
    name = "Unearthly Recovery"
    source = "Sorceror (Divine Soul)"


class EyesOfTheDark(Feature):
    """Starting at lst level, you have darkvision with a range of 120 feet. When
    you reach 3rd level in this class, you learn the darkness spell, which
    doesn’t count against your number of sorcerer spells known. In addition,
    you can cast it by spending 2 sorcery points or by expending a spell
    slot. If you cast it with sorcery points, you can see through the darkness
    created by the spell.

    """
    name = "Eyes of the Dark"
    source = "Sorceror (Shadow Magic)"
    spells_known = (spells.Darkness,)
    spells_prepared = (spells.Darkness,)

    
class StrengthOfTheGrave(Feature):
    """Starting at lst level, your existence in a twilight state between life
    and death makes you difficult to defeat. When damage reduces you to 0 hit
    points, you can make a Charisma saving throw (DC 5 + the damage taken). On
    a success, you instead drop to 1 hit point. You can’t use this feature if
    you are reduced to 0 hit points by radiant damage or by a critical
    hit. After the saving throw succeeds, you can’t use this feature again
    until you finish a long rest

    """
    name = "Strength of the Grave"
    source = "Sorceror (Shadow Magic)"


class HoundOfIllOmen(Feature):
    """At 6th level, you gain the ability to call forth a howling creature of
    darkness to harass your foes. As a bonus action, you can spend 3 sorcery
    points to magically summon a hound of ill omen to target one creature you
    can see within 120 feet of you. The hound uses the dire wolf’s statistics
    (see the Monster Manual or appendix C in the Player’s Handbook), with the
    following changes:

    -- The hound is size Medium, not Large, and it counts as a monstrosity, not
    a beast. 

    -- It appears with a number of temporary hit points equal to half your
    sorcerer level. 

    -- It can move through other creatures and objects as if they were
    difficult terrain. The bound takes 5 force damage if it ends its turn
    inside an object. 

    -- At the start of its turn, the hound automatically knows its target's
    location. If the target was hidden, it is no longer hidden from the hound.
    
    The hound appears in an unoccupied space of your choice within 30 feet of
    the target. Roll initiative for the hound. On its turn, it can move only
    toward its target by the most direct route, and it can use its action only
    to attack its target. The hound can make opportunity attacks but only
    against its target. Additionally, while the hound is within 5 feet of the
    target, the target has disadvantage on saving throws against any spell you
    cast. The hound disappears if it is reduced to 0 hit points, if its target
    is reduced to 0 hit points, or after 5 minutes.

    """
    name = "Hound of Ill Omen"
    source = "Sorceror (Shadow Magic)"


class ShadowWalk(Feature):
    """At 14th level, you gain the ability to step from one shadow into
    another. When you are in dim light or darkness, as a bonus action, you
    can magically teleport up to 120 feet to an unoccupied space you can see
    that is also in dim light or darkness

    """
    name = "Shadow Walk"
    source = "Sorceror (Shadow Magic)"


class UmbralForm(Feature):
    """Starting at 18th level, you can spend 6 sorcery points as a bonus action to
    magically transform yourself into a shadowy form. In this form, you have
    resistance to all damage except force and radiant damage, and you can move
    through other creatures and objects as if they were difficult terrain. You
    take 5 force damage if you end your turn inside an object. You remain in
    this form for 1 minute. It ends early if you are incapacitated, if you die,
    or if you dismiss it as a bonus action.

    """
    name = "Umbral Form"
    source = "Sorceror (Shadow Magic)"

    
# Storm Sorcery
class TempestuousMagic(Feature):
    """Starting at 1st level, you can use a bonus action on your turn to cause
    whirling gusts of elemental air to briefly surround you, immediately before
    or after you cast a spell of 1st level or higher. Doing so allows you to
    fly up to 10 feet without provoking opportunity attacks

    """
    name = "Tempestuous Magic"
    source = "Sorceror (Storm Sorcery)"


class HeartOfTheStorm(Feature):
    """At 6th level, you gain resistance to lightning and thunder damage. In
    addition, whenever you start casting a spell of 1st level or higher that
    deals lightning or thunder damage, stormy magic erupts from you. This
    eruption causes creatures ofyour choice that you can see within 10 feet of
    you to take lightning or thunder damage (choose each time this ability
    activates) equal to half your sorcerer level.

    """
    name = "Heart of the Storm"
    source = "Sorceror (Storm Sorcery)"


class StormGuide(Feature):
    """At 6th level, you gain the ability to subtly control the weather around
    you. Ifit is raining, you can use an action to cause the rain to stop
    falling in a 20-foot—radius sphere centered on you. You can end this effect
    as a bonus action. If it is windy, you can use a bonus action each round to
    choose the direction that the wind blows in a IOO-foot-radius sphere
    centered on you. The wind blows in that direction until the end ofyour next
    turn. This feature doesn't alter the speed of the wind.

    """
    name = "Storm Guide"
    source = "Sorceror (Storm Sorcery)"


class StormsFury(Feature):
    """Starting at 14th level, when you are hit by a melee attack, you can use
    your reaction to deal lightning damage to the attacker. The damage equals
    your sorcerer level. The attacker must also make a Strength saving throw
    against your sorcerer spell save DC. On a failed save, the attacker is
    pushed in a straight line up to 20 feet away from you.

    """
    name = "Storm's Fury"
    source = "Sorceror (Storm Sorcery)"


class WindSoul(Feature):
    """At 18th level, you gain immunity to lightning and thunder damage. You also
    gain a magical flying speed of 60 feet. As an action. you can reduce your
    flying speed to 30 feet for 1 hour and choose a number of creatures within
    30 feet ofyou equal to 3 + your Charisma modifier. The chosen creatures
    gain a magical flying speed of 30 feet for 1 hour. Once you reduce your
    flying speed in this way, you can’t do so again until you finish a short or
    long rest

    """
    name = "Wind Soul"
    source = "Sorceror (Storm Sorcery)"
